"""
Manage LXD containers.

.. versionadded:: 2019.2.0

.. note:

    - :ref:`pylxd` version 2 is required to let this work,
      currently only available via pip.

        To install on Ubuntu:

        $ apt-get install libssl-dev python-pip
        $ pip install -U pylxd

    - you need lxd installed on the minion
      for the init() and version() methods.

    - for the config_get() and config_get() methods
      you need to have lxd-client installed.

.. _pylxd: https://github.com/lxc/pylxd/blob/master/doc/source/installation.rst

:maintainer: René Jochum <rene@jochums.at>
:maturity: new
:depends: python-pylxd
:platform: Linux
"""

from salt.exceptions import CommandExecutionError, SaltInvocationError

__docformat__ = "restructuredtext en"

__virtualname__ = "lxd_container"

# Keep in sync with: https://github.com/lxc/lxd/blob/master/shared/status.go
CONTAINER_STATUS_RUNNING = 103
CONTAINER_STATUS_FROZEN = 110
CONTAINER_STATUS_STOPPED = 102


def __virtual__():
    """
    Only load if the lxd module is available in __salt__
    """
    if "lxd.version" in __salt__:
        return __virtualname__
    return (False, "lxd module could not be loaded")


def present(
    name,
    running=None,
    source=None,
    profiles=None,
    config=None,
    devices=None,
    architecture="x86_64",
    ephemeral=False,
    restart_on_change=False,
    remote_addr=None,
    cert=None,
    key=None,
    verify_cert=True,
):
    """
    Create the named container if it does not exist

    name
        The name of the container to be created

    running : None
        * If ``True``, ensure that the container is running
        * If ``False``, ensure that the container is stopped
        * If ``None``, do nothing with regards to the running state of the
          container

    source : None
        Can be either a string containing an image alias:

        .. code-block:: none

             "xenial/amd64"

        or an dict with type "image" with alias:

        .. code-block:: python

            {"type": "image",
             "alias": "xenial/amd64"}

        or image with "fingerprint":

        .. code-block:: python

            {"type": "image",
             "fingerprint": "SHA-256"}

        or image with "properties":

        .. code-block:: python

            {"type": "image",
             "properties": {
                "os": "ubuntu",
                "release": "14.04",
                "architecture": "x86_64"
             }}

        or none:

        .. code-block:: python

            {"type": "none"}

        or copy:

        .. code-block:: python

            {"type": "copy",
             "source": "my-old-container"}

    profiles : ['default']
        List of profiles to apply on this container

    config :
        A config dict or None (None = unset).

        Can also be a list:

        .. code-block:: python

            [{'key': 'boot.autostart', 'value': 1},
             {'key': 'security.privileged', 'value': '1'}]

    devices :
        A device dict or None (None = unset).

    architecture : 'x86_64'
        Can be one of the following:

        * unknown
        * i686
        * x86_64
        * armv7l
        * aarch64
        * ppc
        * ppc64
        * ppc64le
        * s390x

    ephemeral : False
        Destroy this container after stop?

    restart_on_change : False
        Restart the container when we detect changes on the config or
        its devices?

    remote_addr :
        An URL to a remote Server, you also have to give cert and key if you
        provide remote_addr!

        Examples:
            https://myserver.lan:8443
            /var/lib/mysocket.sock

    cert :
        PEM Formatted SSL Zertifikate.

        Examples:
            ~/.config/lxc/client.crt

    key :
        PEM Formatted SSL Key.

        Examples:
            ~/.config/lxc/client.key

    verify_cert : True
        Wherever to verify the cert, this is by default True
        but in the most cases you want to set it off as LXD
        normally uses self-signed certificates.
    """
    if profiles is None:
        profiles = ["default"]

    if source is None:
        source = {}

    ret = {
        "name": name,
        "running": running,
        "profiles": profiles,
        "source": source,
        "config": config,
        "devices": devices,
        "architecture": architecture,
        "ephemeral": ephemeral,
        "restart_on_change": restart_on_change,
        "remote_addr": remote_addr,
        "cert": cert,
        "key": key,
        "verify_cert": verify_cert,
        "changes": {},
    }

    container = None
    try:
        container = __salt__["lxd.container_get"](
            name, remote_addr, cert, key, verify_cert, _raw=True
        )
    except CommandExecutionError as e:
        return _error(ret, str(e))
    except SaltInvocationError as e:
        # Profile not found
        pass

    if container is None:
        if __opts__["test"]:
            # Test is on, just return that we would create the container
            msg = f'Would create the container "{name}"'
            ret["changes"] = {"created": msg}
            if running is True:
                msg = msg + " and start it."
                ret["changes"]["started"] = 'Would start the container "{}"'.format(
                    name
                )

            ret["changes"] = {"created": msg}
            return _unchanged(ret, msg)

        # create the container
        try:
            __salt__["lxd.container_create"](
                name,
                source,
                profiles,
                config,
                devices,
                architecture,
                ephemeral,
                True,  # Wait
                remote_addr,
                cert,
                key,
                verify_cert,
            )
        except CommandExecutionError as e:
            return _error(ret, str(e))

        msg = f'Created the container "{name}"'
        ret["changes"] = {"created": msg}

        if running is True:
            try:
                __salt__["lxd.container_start"](
                    name, remote_addr, cert, key, verify_cert
                )
            except CommandExecutionError as e:
                return _error(ret, str(e))

            msg = msg + " and started it."
            ret["changes"] = {"started": f'Started the container "{name}"'}

        return _success(ret, msg)

    # Container exists, lets check for differences
    new_profiles = set(map(str, profiles))
    old_profiles = set(map(str, container.profiles))

    container_changed = False

    profile_changes = []
    # Removed profiles
    for k in old_profiles.difference(new_profiles):
        if not __opts__["test"]:
            profile_changes.append(f'Removed profile "{k}"')
            old_profiles.discard(k)
        else:
            profile_changes.append(f'Would remove profile "{k}"')

    # Added profiles
    for k in new_profiles.difference(old_profiles):
        if not __opts__["test"]:
            profile_changes.append(f'Added profile "{k}"')
            old_profiles.add(k)
        else:
            profile_changes.append(f'Would add profile "{k}"')

    if profile_changes:
        container_changed = True
        ret["changes"]["profiles"] = profile_changes
        container.profiles = list(old_profiles)

    # Config and devices changes
    config, devices = __salt__["lxd.normalize_input_values"](config, devices)
    changes = __salt__["lxd.sync_config_devices"](
        container, config, devices, __opts__["test"]
    )
    if changes:
        container_changed = True
        ret["changes"].update(changes)

    is_running = container.status_code == CONTAINER_STATUS_RUNNING

    if not __opts__["test"]:
        try:
            __salt__["lxd.pylxd_save_object"](container)
        except CommandExecutionError as e:
            return _error(ret, str(e))

    if running != is_running:
        if running is True:
            if __opts__["test"]:
                changes["running"] = "Would start the container"
                return _unchanged(
                    ret,
                    f'Container "{name}" would get changed and started.',
                )
            else:
                container.start(wait=True)
                changes["running"] = "Started the container"

        elif running is False:
            if __opts__["test"]:
                changes["stopped"] = "Would stopped the container"
                return _unchanged(
                    ret,
                    f'Container "{name}" would get changed and stopped.',
                )
            else:
                container.stop(wait=True)
                changes["stopped"] = "Stopped the container"

    if (
        (running is True or running is None)
        and is_running
        and restart_on_change
        and container_changed
    ):

        if __opts__["test"]:
            changes["restarted"] = "Would restart the container"
            return _unchanged(ret, f'Would restart the container "{name}"')
        else:
            container.restart(wait=True)
            changes["restarted"] = f'Container "{name}" has been restarted'
            return _success(ret, f'Container "{name}" has been restarted')

    if not container_changed:
        return _success(ret, "No changes")

    if __opts__["test"]:
        return _unchanged(ret, f'Container "{name}" would get changed.')

    return _success(ret, "{} changes".format(len(ret["changes"].keys())))


def absent(name, stop=False, remote_addr=None, cert=None, key=None, verify_cert=True):
    """
    Ensure a LXD container is not present, destroying it if present

    name :
        The name of the container to destroy

    stop :
        stop before destroying
        default: false

    remote_addr :
        An URL to a remote Server, you also have to give cert and key if you
        provide remote_addr!

        Examples:
            https://myserver.lan:8443
            /var/lib/mysocket.sock

    cert :
        PEM Formatted SSL Zertifikate.

        Examples:
            ~/.config/lxc/client.crt

    key :
        PEM Formatted SSL Key.

        Examples:
            ~/.config/lxc/client.key

    verify_cert : True
        Wherever to verify the cert, this is by default True
        but in the most cases you want to set it off as LXD
        normally uses self-signed certificates.
    """
    ret = {
        "name": name,
        "stop": stop,
        "remote_addr": remote_addr,
        "cert": cert,
        "key": key,
        "verify_cert": verify_cert,
        "changes": {},
    }

    try:
        container = __salt__["lxd.container_get"](
            name, remote_addr, cert, key, verify_cert, _raw=True
        )
    except CommandExecutionError as e:
        return _error(ret, str(e))
    except SaltInvocationError as e:
        # Container not found
        return _success(ret, f'Container "{name}" not found.')

    if __opts__["test"]:
        ret["changes"] = {"removed": f'Container "{name}" would get deleted.'}
        return _unchanged(ret, ret["changes"]["removed"])

    if stop and container.status_code == CONTAINER_STATUS_RUNNING:
        container.stop(wait=True)

    container.delete(wait=True)

    ret["changes"]["deleted"] = f'Container "{name}" has been deleted.'
    return _success(ret, ret["changes"]["deleted"])


def running(
    name, restart=False, remote_addr=None, cert=None, key=None, verify_cert=True
):
    """
    Ensure a LXD container is running and restart it if restart is True

    name :
        The name of the container to start/restart.

    restart :
        restart the container if it is already started.

    remote_addr :
        An URL to a remote Server, you also have to give cert and key if you
        provide remote_addr!

        Examples:
            https://myserver.lan:8443
            /var/lib/mysocket.sock

    cert :
        PEM Formatted SSL Zertifikate.

        Examples:
            ~/.config/lxc/client.crt

    key :
        PEM Formatted SSL Key.

        Examples:
            ~/.config/lxc/client.key

    verify_cert : True
        Wherever to verify the cert, this is by default True
        but in the most cases you want to set it off as LXD
        normally uses self-signed certificates.
    """
    ret = {
        "name": name,
        "restart": restart,
        "remote_addr": remote_addr,
        "cert": cert,
        "key": key,
        "verify_cert": verify_cert,
        "changes": {},
    }

    try:
        container = __salt__["lxd.container_get"](
            name, remote_addr, cert, key, verify_cert, _raw=True
        )
    except CommandExecutionError as e:
        return _error(ret, str(e))
    except SaltInvocationError as e:
        # Container not found
        return _error(ret, f'Container "{name}" not found')

    is_running = container.status_code == CONTAINER_STATUS_RUNNING

    if is_running:
        if not restart:
            return _success(ret, f'The container "{name}" is already running')
        else:
            if __opts__["test"]:
                ret["changes"]["restarted"] = 'Would restart the container "{}"'.format(
                    name
                )
                return _unchanged(ret, ret["changes"]["restarted"])
            else:
                container.restart(wait=True)
                ret["changes"]["restarted"] = 'Restarted the container "{}"'.format(
                    name
                )
                return _success(ret, ret["changes"]["restarted"])

    if __opts__["test"]:
        ret["changes"]["started"] = f'Would start the container "{name}"'
        return _unchanged(ret, ret["changes"]["started"])

    container.start(wait=True)
    ret["changes"]["started"] = f'Started the container "{name}"'
    return _success(ret, ret["changes"]["started"])


def frozen(name, start=True, remote_addr=None, cert=None, key=None, verify_cert=True):
    """
    Ensure a LXD container is frozen, start and freeze it if start is true

    name :
        The name of the container to freeze

    start :
        start and freeze it

    remote_addr :
        An URL to a remote Server, you also have to give cert and key if you
        provide remote_addr!

        Examples:
            https://myserver.lan:8443
            /var/lib/mysocket.sock

    cert :
        PEM Formatted SSL Zertifikate.

        Examples:
            ~/.config/lxc/client.crt

    key :
        PEM Formatted SSL Key.

        Examples:
            ~/.config/lxc/client.key

    verify_cert : True
        Wherever to verify the cert, this is by default True
        but in the most cases you want to set it off as LXD
        normally uses self-signed certificates.
    """
    ret = {
        "name": name,
        "start": start,
        "remote_addr": remote_addr,
        "cert": cert,
        "key": key,
        "verify_cert": verify_cert,
        "changes": {},
    }

    try:
        container = __salt__["lxd.container_get"](
            name, remote_addr, cert, key, verify_cert, _raw=True
        )
    except CommandExecutionError as e:
        return _error(ret, str(e))
    except SaltInvocationError as e:
        # Container not found
        return _error(ret, f'Container "{name}" not found')

    if container.status_code == CONTAINER_STATUS_FROZEN:
        return _success(ret, f'Container "{name}" is alredy frozen')

    is_running = container.status_code == CONTAINER_STATUS_RUNNING

    if not is_running and not start:
        return _error(
            ret,
            'Container "{}" is not running and start is False, cannot freeze it'.format(
                name
            ),
        )

    elif not is_running and start:
        if __opts__["test"]:
            ret["changes"][
                "started"
            ] = f'Would start the container "{name}" and freeze it after'
            return _unchanged(ret, ret["changes"]["started"])
        else:
            container.start(wait=True)
            ret["changes"]["started"] = f'Start the container "{name}"'

    if __opts__["test"]:
        ret["changes"]["frozen"] = f'Would freeze the container "{name}"'
        return _unchanged(ret, ret["changes"]["frozen"])

    container.freeze(wait=True)
    ret["changes"]["frozen"] = f'Froze the container "{name}"'

    return _success(ret, ret["changes"]["frozen"])


def stopped(name, kill=False, remote_addr=None, cert=None, key=None, verify_cert=True):
    """
    Ensure a LXD container is stopped, kill it if kill is true else stop it

    name :
        The name of the container to stop

    kill :
        kill if true

    remote_addr :
        An URL to a remote Server, you also have to give cert and key if you
        provide remote_addr!

        Examples:
            https://myserver.lan:8443
            /var/lib/mysocket.sock

    cert :
        PEM Formatted SSL Zertifikate.

        Examples:
            ~/.config/lxc/client.crt

    key :
        PEM Formatted SSL Key.

        Examples:
            ~/.config/lxc/client.key

    verify_cert : True
        Wherever to verify the cert, this is by default True
        but in the most cases you want to set it off as LXD
        normally uses self-signed certificates.
    """
    ret = {
        "name": name,
        "kill": kill,
        "remote_addr": remote_addr,
        "cert": cert,
        "key": key,
        "verify_cert": verify_cert,
        "changes": {},
    }

    try:
        container = __salt__["lxd.container_get"](
            name, remote_addr, cert, key, verify_cert, _raw=True
        )
    except CommandExecutionError as e:
        return _error(ret, str(e))
    except SaltInvocationError as e:
        # Container not found
        return _error(ret, f'Container "{name}" not found')

    if container.status_code == CONTAINER_STATUS_STOPPED:
        return _success(ret, f'Container "{name}" is already stopped')

    if __opts__["test"]:
        ret["changes"]["stopped"] = f'Would stop the container "{name}"'
        return _unchanged(ret, ret["changes"]["stopped"])

    container.stop(force=kill, wait=True)
    ret["changes"]["stopped"] = f'Stopped the container "{name}"'
    return _success(ret, ret["changes"]["stopped"])


def migrated(
    name,
    remote_addr,
    cert,
    key,
    verify_cert,
    src_remote_addr,
    stop_and_start=False,
    src_cert=None,
    src_key=None,
    src_verify_cert=None,
):
    """Ensure a container is migrated to another host

    If the container is running, it either must be shut down
    first (use stop_and_start=True) or criu must be installed
    on the source and destination machines.

    For this operation both certs need to be authenticated,
    use :mod:`lxd.authenticate <salt.states.lxd.authenticate`
    to authenticate your cert(s).

    name :
        The container to migrate

    remote_addr :
        An URL to the destination remote Server

        Examples:
            https://myserver.lan:8443
            /var/lib/mysocket.sock

    cert :
        PEM Formatted SSL Zertifikate.

        Examples:
            ~/.config/lxc/client.crt

    key :
        PEM Formatted SSL Key.

        Examples:
            ~/.config/lxc/client.key

    verify_cert : True
        Wherever to verify the cert, this is by default True
        but in the most cases you want to set it off as LXD
        normally uses self-signed certificates.

    src_remote_addr :
        An URL to the source remote Server

        Examples:
            https://myserver.lan:8443
            /var/lib/mysocket.sock

    stop_and_start:
        Stop before migrating and start after

    src_cert :
        PEM Formatted SSL Zertifikate, if None we copy "cert"

        Examples:
            ~/.config/lxc/client.crt

    src_key :
        PEM Formatted SSL Key, if None we copy "key"

        Examples:
            ~/.config/lxc/client.key

    src_verify_cert :
        Wherever to verify the cert, if None we copy "verify_cert"
    """
    ret = {
        "name": name,
        "remote_addr": remote_addr,
        "cert": cert,
        "key": key,
        "verify_cert": verify_cert,
        "src_remote_addr": src_remote_addr,
        "src_and_start": stop_and_start,
        "src_cert": src_cert,
        "src_key": src_key,
        "changes": {},
    }

    dest_container = None
    try:
        dest_container = __salt__["lxd.container_get"](
            name, remote_addr, cert, key, verify_cert, _raw=True
        )
    except CommandExecutionError as e:
        return _error(ret, str(e))
    except SaltInvocationError as e:
        # Destination container not found
        pass

    if dest_container is not None:
        return _success(ret, f'Container "{name}" exists on the destination')

    if src_verify_cert is None:
        src_verify_cert = verify_cert

    try:
        __salt__["lxd.container_get"](
            name, src_remote_addr, src_cert, src_key, src_verify_cert, _raw=True
        )
    except CommandExecutionError as e:
        return _error(ret, str(e))
    except SaltInvocationError as e:
        # Container not found
        return _error(ret, f'Source Container "{name}" not found')

    if __opts__["test"]:
        ret["changes"]["migrated"] = (
            'Would migrate the container "{}" from "{}" to "{}"'.format(
                name, src_remote_addr, remote_addr
            )
        )
        return _unchanged(ret, ret["changes"]["migrated"])

    try:
        __salt__["lxd.container_migrate"](
            name,
            stop_and_start,
            remote_addr,
            cert,
            key,
            verify_cert,
            src_remote_addr,
            src_cert,
            src_key,
            src_verify_cert,
        )
    except CommandExecutionError as e:
        return _error(ret, str(e))

    ret["changes"]["migrated"] = 'Migrated the container "{}" from "{}" to "{}"'.format(
        name, src_remote_addr, remote_addr
    )
    return _success(ret, ret["changes"]["migrated"])


def _success(ret, success_msg):
    ret["result"] = True
    ret["comment"] = success_msg
    if "changes" not in ret:
        ret["changes"] = {}
    return ret


def _unchanged(ret, msg):
    ret["result"] = None
    ret["comment"] = msg
    if "changes" not in ret:
        ret["changes"] = {}
    return ret


def _error(ret, err_msg):
    ret["result"] = False
    ret["comment"] = err_msg
    if "changes" not in ret:
        ret["changes"] = {}
    return ret



"""
Debian Package builder system

.. versionadded:: 2015.8.0

This system allows for all of the components to build debs safely in chrooted
environments. This also provides a function to generate debian repositories

This module implements the pkgbuild interface
"""

import errno
import logging
import os
import re
import shutil
import tempfile
import time
import traceback
import urllib.parse

import salt.utils.files
import salt.utils.path
import salt.utils.stringutils
import salt.utils.vt
from salt.exceptions import CommandExecutionError, SaltInvocationError

HAS_LIBS = False

SIGN_PROMPT_RE = re.compile(r"Enter passphrase: ", re.M)
REPREPRO_SIGN_PROMPT_RE = re.compile(r"Passphrase: ", re.M)

try:
    import gnupg  # pylint: disable=unused-import

    import salt.modules.gpg

    HAS_LIBS = True
except ImportError:
    pass

log = logging.getLogger(__name__)


__virtualname__ = "pkgbuild"


def __virtual__():
    """
    Confirm this module is on a Debian-based system, and has required utilities
    """
    if __grains__.get("os_family", False) in ("Kali", "Debian"):
        missing_util = False
        utils_reqd = ["gpg", "debuild", "pbuilder", "reprepro"]
        for named_util in utils_reqd:
            if not salt.utils.path.which(named_util):
                missing_util = True
                break
        if HAS_LIBS and not missing_util:
            return __virtualname__
        else:
            return (
                False,
                "The debbuild module could not be loaded: requires python-gnupg, gpg,"
                " debuild, pbuilder and reprepro utilities to be installed",
            )
    else:
        return (False, "The debbuild module could not be loaded: unsupported OS family")


def _check_repo_sign_utils_support(name):
    """
    Check for specified command name in search path
    """
    if salt.utils.path.which(name):
        return True
    else:
        raise CommandExecutionError(
            "utility '{}' needs to be installed or made available in search path".format(
                name
            )
        )


def _check_repo_gpg_phrase_utils():
    """
    Check for /usr/lib/gnupg2/gpg-preset-passphrase is installed
    """
    util_name = "/usr/lib/gnupg2/gpg-preset-passphrase"
    if __salt__["file.file_exists"](util_name):
        return True
    else:
        raise CommandExecutionError(f"utility '{util_name}' needs to be installed")


def _get_build_env(env):
    """
    Get build environment overrides dictionary to use in build process
    """
    env_override = ""
    if env is None:
        return env_override
    if not isinstance(env, dict):
        raise SaltInvocationError("'env' must be a Python dictionary")
    for key, value in env.items():
        env_override += f"{key}={value}\n"
        env_override += f"export {key}\n"
    return env_override


def _get_repo_options_env(env):
    """
    Get repo environment overrides dictionary to use in repo options process

    env
        A dictionary of variables to define the repository options
        Example:

        .. code-block:: yaml

            - env:
                - OPTIONS : 'ask-passphrase'

        .. warning::

            The above illustrates a common PyYAML pitfall, that **yes**,
            **no**, **on**, **off**, **true**, and **false** are all loaded as
            boolean ``True`` and ``False`` values, and must be enclosed in
            quotes to be used as strings. More info on this (and other) PyYAML
            idiosyncrasies can be found :ref:`here <yaml-idiosyncrasies>`.

    """
    env_options = ""
    if env is None:
        return env_options
    if not isinstance(env, dict):
        raise SaltInvocationError("'env' must be a Python dictionary")
    for key, value in env.items():
        if key == "OPTIONS":
            env_options += f"{value}\n"
    return env_options


def _get_repo_dists_env(env):
    """
    Get repo environment overrides dictionary to use in repo distributions process

    env
        A dictionary of variables to define the repository distributions
        Example:

        .. code-block:: yaml

            - env:
                - ORIGIN : 'jessie'
                - LABEL : 'salt debian'
                - SUITE : 'main'
                - VERSION : '8.1'
                - CODENAME : 'jessie'
                - ARCHS : 'amd64 i386 source'
                - COMPONENTS : 'main'
                - DESCRIPTION : 'SaltStack Debian package repo'

        .. warning::

            The above illustrates a common PyYAML pitfall, that **yes**,
            **no**, **on**, **off**, **true**, and **false** are all loaded as
            boolean ``True`` and ``False`` values, and must be enclosed in
            quotes to be used as strings. More info on this (and other) PyYAML
            idiosyncrasies can be found :ref:`here <yaml-idiosyncrasies>`.

    """
    # env key with tuple of control information for handling input env dictionary
    # 0 | M - Mandatory, O - Optional, I - Ignore
    # 1 | 'text string for repo field'
    # 2 | 'default value'
    dflts_dict = {
        "OPTIONS": ("I", "", "processed by _get_repo_options_env"),
        "ORIGIN": ("O", "Origin", "SaltStack"),
        "LABEL": ("O", "Label", "salt_debian"),
        "SUITE": ("O", "Suite", "stable"),
        "VERSION": ("O", "Version", "9.0"),
        "CODENAME": ("M", "Codename", "stretch"),
        "ARCHS": ("M", "Architectures", "i386 amd64 source"),
        "COMPONENTS": ("M", "Components", "main"),
        "DESCRIPTION": ("O", "Description", "SaltStack debian package repo"),
    }

    env_dists = ""
    codename = ""
    dflts_keys = list(dflts_dict.keys())
    if env is None:
        for key, value in dflts_dict.items():
            if dflts_dict[key][0] == "M":
                env_dists += f"{dflts_dict[key][1]}: {dflts_dict[key][2]}\n"
                if key == "CODENAME":
                    codename = dflts_dict[key][2]
        return (codename, env_dists)

    if not isinstance(env, dict):
        raise SaltInvocationError("'env' must be a Python dictionary")

    env_man_seen = []
    for key, value in env.items():
        if key in dflts_keys:
            if dflts_dict[key][0] == "M":
                env_man_seen.append(key)
                if key == "CODENAME":
                    codename = value
            if dflts_dict[key][0] != "I":
                env_dists += f"{dflts_dict[key][1]}: {value}\n"
        else:
            env_dists += f"{key}: {value}\n"

    # ensure mandatories are included
    env_keys = list(env.keys())
    for key in env_keys:
        if key in dflts_keys and dflts_dict[key][0] == "M" and key not in env_man_seen:
            env_dists += f"{dflts_dict[key][1]}: {dflts_dict[key][2]}\n"
            if key == "CODENAME":
                codename = value

    return (codename, env_dists)


def _create_pbuilders(env, runas="root"):
    """
    Create the .pbuilder family of files in user's home directory

    env
        A list  or dictionary of environment variables to be set prior to execution.
        Example:

        .. code-block:: yaml

            - env:
                - DEB_BUILD_OPTIONS: 'nocheck'

        .. warning::

            The above illustrates a common PyYAML pitfall, that **yes**,
            **no**, **on**, **off**, **true**, and **false** are all loaded as
            boolean ``True`` and ``False`` values, and must be enclosed in
            quotes to be used as strings. More info on this (and other) PyYAML
            idiosyncrasies can be found :ref:`here <yaml-idiosyncrasies>`.

    runas : root
        .. versionadded:: 2019.2.1

        User to create the files and directories

        .. note::

            Ensure the user has correct permissions to any files and
            directories which are to be utilized.
    """
    home = os.path.expanduser(f"~{runas}")
    pbuilderrc = os.path.join(home, ".pbuilderrc")
    if not os.path.isfile(pbuilderrc):
        raise SaltInvocationError("pbuilderrc environment is incorrectly setup")

    env_overrides = _get_build_env(env)
    if env_overrides and not env_overrides.isspace():
        with salt.utils.files.fopen(pbuilderrc, "a") as fow:
            fow.write(salt.utils.stringutils.to_str(env_overrides))
    cmd = "chown {0}:{0} {1}".format(runas, pbuilderrc)
    retrc = __salt__["cmd.retcode"](cmd, runas="root")
    if retrc != 0:
        raise SaltInvocationError(
            "Create pbuilderrc in home directory failed with return error '{}', "
            "check logs for further details".format(retrc)
        )


def _mk_tree():
    """
    Create the debian build area
    """
    basedir = tempfile.mkdtemp()
    return basedir


def _get_spec(tree_base, spec, saltenv="base"):
    """
    Get the spec file (tarball of the debian sub-dir to use)
    and place it in build area

    """
    spec_tgt = os.path.basename(spec)
    dest = os.path.join(tree_base, spec_tgt)
    return __salt__["cp.get_url"](spec, dest, saltenv=saltenv)


def _get_src(tree_base, source, saltenv="base"):
    """
    Get the named sources and place them into the tree_base
    """
    parsed = urllib.parse.urlparse(source)
    sbase = os.path.basename(source)
    dest = os.path.join(tree_base, sbase)
    if parsed.scheme:
        __salt__["cp.get_url"](source, dest, saltenv=saltenv)
    else:
        shutil.copy(source, dest)


def make_src_pkg(dest_dir, spec, sources, env=None, saltenv="base", runas="root"):
    """
    Create a platform specific source package from the given platform spec/control file and sources

    CLI Example:

    **Debian**

    .. code-block:: bash

        salt '*' pkgbuild.make_src_pkg /var/www/html/
                https://raw.githubusercontent.com/saltstack/libnacl/master/pkg/deb/python-libnacl.control.tar.xz
                https://pypi.python.org/packages/source/l/libnacl/libnacl-1.3.5.tar.gz

    This example command should build the libnacl SOURCE package and place it in
    /var/www/html/ on the minion

    dest_dir
        Absolute path for directory to write source package

    spec
        Absolute path to spec file or equivalent

    sources
        Absolute path to source files to build source package from

    env : None
        A list  or dictionary of environment variables to be set prior to execution.
        Example:

        .. code-block:: yaml

            - env:
                - DEB_BUILD_OPTIONS: 'nocheck'

        .. warning::

            The above illustrates a common PyYAML pitfall, that **yes**,
            **no**, **on**, **off**, **true**, and **false** are all loaded as
            boolean ``True`` and ``False`` values, and must be enclosed in
            quotes to be used as strings. More info on this (and other) PyYAML
            idiosyncrasies can be found :ref:`here <yaml-idiosyncrasies>`.

    saltenv: base

        Salt environment variables


    runas : root
        .. versionadded:: 2019.2.1

        User to create the files and directories

        .. note::

            Ensure the user has correct permissions to any files and
            directories which are to be utilized.
    """
    _create_pbuilders(env, runas)
    tree_base = _mk_tree()
    ret = []
    if not os.path.isdir(dest_dir):
        os.makedirs(dest_dir)

    # ensure directories are writable
    root_user = "root"
    retrc = 0
    cmd = "chown {0}:{0} {1}".format(runas, tree_base)
    retrc = __salt__["cmd.retcode"](cmd, runas="root")
    if retrc != 0:
        raise SaltInvocationError(
            "make_src_pkg ensuring tree_base '{}' ownership failed with return error"
            " '{}', check logs for further details".format(tree_base, retrc)
        )

    cmd = "chown {0}:{0} {1}".format(runas, dest_dir)
    retrc = __salt__["cmd.retcode"](cmd, runas=root_user)
    if retrc != 0:
        raise SaltInvocationError(
            "make_src_pkg ensuring dest_dir '{}' ownership failed with return error"
            " '{}', check logs for further details".format(dest_dir, retrc)
        )

    spec_pathfile = _get_spec(tree_base, spec, saltenv)

    # build salt equivalents from scratch
    if isinstance(sources, str):
        sources = sources.split(",")
    for src in sources:
        _get_src(tree_base, src, saltenv)

    # .dsc then assumes sources already build
    if spec_pathfile.endswith(".dsc"):
        for efile in os.listdir(tree_base):
            full = os.path.join(tree_base, efile)
            trgt = os.path.join(dest_dir, efile)
            shutil.copy(full, trgt)
            ret.append(trgt)

        return ret

    # obtain name of 'sdist' generated tarball, extract the version
    # and manipulate the name for debian use (convert minix and add '+ds')
    salttarball = None
    for afile in os.listdir(tree_base):
        if afile.startswith("salt-") and afile.endswith(".tar.gz"):
            salttarball = afile
            break
    else:
        return ret

    frontname = salttarball.split(".tar.gz")
    salttar_name = frontname[0]
    k = salttar_name.rfind("-")
    debname = salttar_name[:k] + "_" + salttar_name[k + 1 :]
    debname += "+ds"
    debname_orig = debname + ".orig.tar.gz"
    abspath_debname = os.path.join(tree_base, debname)

    cmd = f"tar -xvzf {salttarball}"
    retrc = __salt__["cmd.retcode"](cmd, cwd=tree_base, runas=root_user)
    cmd = f"mv {salttar_name} {debname}"
    retrc |= __salt__["cmd.retcode"](cmd, cwd=tree_base, runas=root_user)
    cmd = f"tar -cvzf {os.path.join(tree_base, debname_orig)} {debname}"
    retrc |= __salt__["cmd.retcode"](cmd, cwd=tree_base, runas=root_user)
    cmd = f"rm -f {salttarball}"
    retrc |= __salt__["cmd.retcode"](cmd, cwd=tree_base, runas=root_user, env=env)
    cmd = f"cp {spec_pathfile}  {abspath_debname}"
    retrc |= __salt__["cmd.retcode"](cmd, cwd=abspath_debname, runas=root_user)
    cmd = f"tar -xvJf {spec_pathfile}"
    retrc |= __salt__["cmd.retcode"](cmd, cwd=abspath_debname, runas=root_user, env=env)
    cmd = f"rm -f {os.path.basename(spec_pathfile)}"
    retrc |= __salt__["cmd.retcode"](cmd, cwd=abspath_debname, runas=root_user)
    cmd = "debuild -S -uc -us -sa"
    retrc |= __salt__["cmd.retcode"](
        cmd, cwd=abspath_debname, runas=root_user, python_shell=True, env=env
    )
    cmd = f"rm -fR {abspath_debname}"
    retrc |= __salt__["cmd.retcode"](cmd, runas=root_user)
    if retrc != 0:
        raise SaltInvocationError(
            "Make source package for destination directory {}, spec {}, sources {},"
            " failed with return error {}, check logs for further details".format(
                dest_dir, spec, sources, retrc
            )
        )

    for dfile in os.listdir(tree_base):
        if not dfile.endswith(".build"):
            full = os.path.join(tree_base, dfile)
            trgt = os.path.join(dest_dir, dfile)
            shutil.copy(full, trgt)
            ret.append(trgt)

    return ret


def build(
    runas,
    tgt,
    dest_dir,
    spec,
    sources,
    deps,
    env,
    template,
    saltenv="base",
    log_dir="/var/log/salt/pkgbuild",
):  # pylint: disable=unused-argument
    """
    Given the package destination directory, the tarball containing debian files (e.g. control)
    and package sources, use pbuilder to safely build the platform package

    CLI Example:

    **Debian**

    .. code-block:: bash

        salt '*' pkgbuild.make_src_pkg deb-8-x86_64 /var/www/html
                https://raw.githubusercontent.com/saltstack/libnacl/master/pkg/deb/python-libnacl.control
                https://pypi.python.org/packages/source/l/libnacl/libnacl-1.3.5.tar.gz

    This example command should build the libnacl package for Debian using pbuilder
    and place it in /var/www/html/ on the minion
    """
    ret = {}
    retrc = 0
    try:
        os.makedirs(dest_dir)
    except OSError as exc:
        if exc.errno != errno.EEXIST:
            raise
    dsc_dir = tempfile.mkdtemp()
    try:
        dscs = make_src_pkg(dsc_dir, spec, sources, env, saltenv, runas)
    except Exception as exc:  # pylint: disable=broad-except
        shutil.rmtree(dsc_dir)
        log.error("Failed to make src package, exception '%s'", exc)
        return ret

    root_user = "root"

    # ensure pbuilder setup from runas if other than root
    if runas != root_user:
        user_home = os.path.expanduser(f"~{runas}")
        root_home = os.path.expanduser("~root")
        cmd = f"cp {user_home}/.pbuilderrc {root_home}/"
        retrc = __salt__["cmd.retcode"](
            cmd, runas=root_user, python_shell=True, env=env
        )
        cmd = f"cp -R {user_home}/.pbuilder-hooks {root_home}/"
        retrc = __salt__["cmd.retcode"](
            cmd, runas=root_user, python_shell=True, env=env
        )
        if retrc != 0:
            raise SaltInvocationError(
                "build copy pbuilder files from '{}' to '{}' returned error '{}', "
                "check logs for further details".format(user_home, root_home, retrc)
            )

    cmd = "/usr/sbin/pbuilder --create"
    retrc = __salt__["cmd.retcode"](cmd, runas=root_user, python_shell=True, env=env)
    if retrc != 0:
        raise SaltInvocationError(
            "pbuilder create failed with return error '{}', "
            "check logs for further details".format(retrc)
        )

    # use default /var/cache/pbuilder/result
    results_dir = "/var/cache/pbuilder/result"

    # ensure clean
    cmd = f"rm -fR {results_dir}"
    retrc |= __salt__["cmd.retcode"](cmd, runas=root_user, python_shell=True, env=env)

    # dscs should only contain salt orig and debian tarballs and dsc file
    for dsc in dscs:
        afile = os.path.basename(dsc)
        os.path.join(dest_dir, afile)

        if dsc.endswith(".dsc"):
            dbase = os.path.dirname(dsc)
            try:
                cmd = "chown {0}:{0} -R {1}".format(runas, dbase)
                retrc |= __salt__["cmd.retcode"](
                    cmd, runas=root_user, python_shell=True, env=env
                )
                cmd = "/usr/sbin/pbuilder update --override-config"
                retrc |= __salt__["cmd.retcode"](
                    cmd, runas=root_user, python_shell=True, env=env
                )
                cmd = f'/usr/sbin/pbuilder build --debbuildopts "-sa" {dsc}'
                retrc |= __salt__["cmd.retcode"](
                    cmd, runas=root_user, python_shell=True, env=env
                )
                if retrc != 0:
                    raise SaltInvocationError(
                        "pbuilder build or update failed with return error {}, "
                        "check logs for further details".format(retrc)
                    )

                # ignore local deps generated package file
                for bfile in os.listdir(results_dir):
                    if bfile != "Packages":
                        full = os.path.join(results_dir, bfile)
                        bdist = os.path.join(dest_dir, bfile)
                        shutil.copy(full, bdist)
                        ret.setdefault("Packages", []).append(bdist)

            except Exception as exc:  # pylint: disable=broad-except
                log.error("Error building from '%s', execption '%s'", dsc, exc)

    # remove any Packages file created for local dependency processing
    for pkgzfile in os.listdir(dest_dir):
        if pkgzfile == "Packages":
            pkgzabsfile = os.path.join(dest_dir, pkgzfile)
            os.remove(pkgzabsfile)

    cmd = "chown {0}:{0} -R {1}".format(runas, dest_dir)
    __salt__["cmd.retcode"](cmd, runas=root_user, python_shell=True, env=env)

    shutil.rmtree(dsc_dir)
    return ret


def make_repo(
    repodir,
    keyid=None,
    env=None,
    use_passphrase=False,
    gnupghome="/etc/salt/gpgkeys",
    runas="root",
    timeout=15.0,
):
    """
    Make a package repository and optionally sign it and packages present

    Given the repodir (directory to create repository in), create a Debian
    repository and optionally sign it and packages present. This state is
    best used with onchanges linked to your package building states.

    repodir
        The directory to find packages that will be in the repository.

    keyid
        .. versionchanged:: 2016.3.0

        Optional Key ID to use in signing packages and repository.
        This consists of the last 8 hex digits of the GPG key ID.

        Utilizes Public and Private keys associated with keyid which have
        been loaded into the minion's Pillar data. Leverages gpg-agent and
        gpg-preset-passphrase for caching keys, etc.
        These pillar values are assumed to be filenames which are present
        in ``gnupghome``. The pillar keys shown below have to match exactly.

        For example, contents from a Pillar data file with named Public
        and Private keys as follows:

        .. code-block:: yaml

            gpg_pkg_priv_keyname: gpg_pkg_key.pem
            gpg_pkg_pub_keyname: gpg_pkg_key.pub

    env
        .. versionchanged:: 2016.3.0

        A dictionary of environment variables to be utilized in creating the
        repository.

    use_passphrase : False
        .. versionadded:: 2016.3.0

        Use a passphrase with the signing key presented in ``keyid``.
        Passphrase is received from Pillar data which could be passed on the
        command line with ``pillar`` parameter. For example:

        .. code-block:: bash

            pillar='{ "gpg_passphrase" : "my_passphrase" }'

    gnupghome : /etc/salt/gpgkeys
        .. versionadded:: 2016.3.0

        Location where GPG related files are stored, used with ``keyid``.

    runas : root
        .. versionadded:: 2016.3.0

        User to create the repository as, and optionally sign packages.

        .. note::

            Ensure the user has correct permissions to any files and
            directories which are to be utilized.

    timeout : 15.0
        .. versionadded:: 2016.3.4

        Timeout in seconds to wait for the prompt for inputting the passphrase.

    CLI Example:

    .. code-block:: bash

        salt '*' pkgbuild.make_repo /var/www/html

    """
    res = {"retcode": 1, "stdout": "", "stderr": "initialization value"}

    retrc = 0

    if gnupghome and env is None:
        env = {}
        env["GNUPGHOME"] = gnupghome

    repoconf = os.path.join(repodir, "conf")
    if not os.path.isdir(repoconf):
        os.makedirs(repoconf)

    codename, repocfg_dists = _get_repo_dists_env(env)
    repoconfdist = os.path.join(repoconf, "distributions")
    with salt.utils.files.fopen(repoconfdist, "w") as fow:
        fow.write(salt.utils.stringutils.to_str(repocfg_dists))

    repocfg_opts = _get_repo_options_env(env)
    repoconfopts = os.path.join(repoconf, "options")
    with salt.utils.files.fopen(repoconfopts, "w") as fow:
        fow.write(salt.utils.stringutils.to_str(repocfg_opts))

    cmd = "chown {0}:{0} -R {1}".format(runas, repoconf)
    retrc = __salt__["cmd.retcode"](cmd, runas="root")
    if retrc != 0:
        raise SaltInvocationError(
            "failed to ensure rights to repoconf directory, error {}, "
            "check logs for further details".format(retrc)
        )

    local_keygrip_to_use = None
    local_key_fingerprint = None
    local_keyid = None
    phrase = ""

    # preset passphase and interaction with gpg-agent
    gpg_info_file = f"{gnupghome}/gpg-agent-info-salt"
    gpg_tty_info_file = f"{gnupghome}/gpg-tty-info-salt"

    # if using older than gnupg 2.1, then env file exists
    older_gnupg = __salt__["file.file_exists"](gpg_info_file)

    if keyid is not None:
        with salt.utils.files.fopen(repoconfdist, "a") as fow:
            fow.write(salt.utils.stringutils.to_str(f"SignWith: {keyid}\n"))

        # import_keys
        pkg_pub_key_file = "{}/{}".format(
            gnupghome, __salt__["pillar.get"]("gpg_pkg_pub_keyname", None)
        )
        pkg_priv_key_file = "{}/{}".format(
            gnupghome, __salt__["pillar.get"]("gpg_pkg_priv_keyname", None)
        )

        if pkg_pub_key_file is None or pkg_priv_key_file is None:
            raise SaltInvocationError(
                "Pillar data should contain Public and Private keys associated with"
                " 'keyid'"
            )
        try:
            __salt__["gpg.import_key"](
                user=runas, filename=pkg_pub_key_file, gnupghome=gnupghome
            )
            __salt__["gpg.import_key"](
                user=runas, filename=pkg_priv_key_file, gnupghome=gnupghome
            )

        except SaltInvocationError:
            raise SaltInvocationError(
                "Public and Private key files associated with Pillar data and 'keyid' "
                "{} could not be found".format(keyid)
            )

        # gpg keys should have been loaded as part of setup
        # retrieve specified key, obtain fingerprint and preset passphrase
        local_keys = __salt__["gpg.list_keys"](user=runas, gnupghome=gnupghome)
        for gpg_key in local_keys:
            if keyid == gpg_key["keyid"][8:]:
                local_keygrip_to_use = gpg_key["fingerprint"]
                local_key_fingerprint = gpg_key["fingerprint"]
                local_keyid = gpg_key["keyid"]
                break

        if not older_gnupg:
            try:
                _check_repo_sign_utils_support("gpg2")
                cmd = "gpg2 --with-keygrip --list-secret-keys"
            except CommandExecutionError:
                # later gpg versions have dispensed with gpg2 - Ubuntu 18.04
                cmd = "gpg --with-keygrip --list-secret-keys"
            local_keys2_keygrip = __salt__["cmd.run"](cmd, runas=runas, env=env)
            local_keys2 = iter(local_keys2_keygrip.splitlines())
            try:
                for line in local_keys2:
                    if line.startswith("sec"):
                        line_fingerprint = next(local_keys2).lstrip().rstrip()
                        if local_key_fingerprint == line_fingerprint:
                            lkeygrip = next(local_keys2).split("=")
                            local_keygrip_to_use = lkeygrip[1].lstrip().rstrip()
                            break
            except StopIteration:
                raise SaltInvocationError(
                    "unable to find keygrip associated with fingerprint '{}' for keyid"
                    " '{}'".format(local_key_fingerprint, local_keyid)
                )

        if local_keyid is None:
            raise SaltInvocationError(
                "The key ID '{}' was not found in GnuPG keyring at '{}'".format(
                    keyid, gnupghome
                )
            )

        _check_repo_sign_utils_support("debsign")

        if older_gnupg:
            with salt.utils.files.fopen(gpg_info_file, "r") as fow:
                gpg_raw_info = fow.readlines()

            for gpg_info_line in gpg_raw_info:
                gpg_info_line = salt.utils.stringutils.to_unicode(gpg_info_line)
                gpg_info = gpg_info_line.split("=")
                env[gpg_info[0]] = gpg_info[1]
                break
        else:
            with salt.utils.files.fopen(gpg_tty_info_file, "r") as fow:
                gpg_raw_info = fow.readlines()

            for gpg_tty_info_line in gpg_raw_info:
                gpg_tty_info_line = salt.utils.stringutils.to_unicode(gpg_tty_info_line)
                gpg_tty_info = gpg_tty_info_line.split("=")
                env[gpg_tty_info[0]] = gpg_tty_info[1]
                break

        if use_passphrase:
            _check_repo_gpg_phrase_utils()
            phrase = __salt__["pillar.get"]("gpg_passphrase")
            cmd = (
                "/usr/lib/gnupg2/gpg-preset-passphrase --verbose --preset --passphrase"
                ' "{}" {}'.format(phrase, local_keygrip_to_use)
            )
            retrc |= __salt__["cmd.retcode"](cmd, runas=runas, env=env)

    for debfile in os.listdir(repodir):
        abs_file = os.path.join(repodir, debfile)
        if debfile.endswith(".changes"):
            os.remove(abs_file)

        if debfile.endswith(".dsc"):
            # sign_it_here
            if older_gnupg:
                if local_keyid is not None:
                    cmd = f"debsign --re-sign -k {keyid} {abs_file}"
                    retrc |= __salt__["cmd.retcode"](
                        cmd, runas=runas, cwd=repodir, use_vt=True, env=env
                    )

                cmd = (
                    "reprepro --ignore=wrongdistribution --component=main -Vb ."
                    " includedsc {} {}".format(codename, abs_file)
                )
                retrc |= __salt__["cmd.retcode"](
                    cmd, runas=runas, cwd=repodir, use_vt=True, env=env
                )
            else:
                # interval of 0.125 is really too fast on some systems
                interval = 0.5
                if local_keyid is not None:
                    number_retries = timeout / interval
                    times_looped = 0
                    error_msg = f"Failed to debsign file {abs_file}"
                    if (
                        __grains__["os"] in ["Ubuntu"]
                        and __grains__["osmajorrelease"] < 18
                    ) or (
                        __grains__["os"] in ["Debian"]
                        and __grains__["osmajorrelease"] <= 8
                    ):
                        cmd = f"debsign --re-sign -k {keyid} {abs_file}"
                        try:
                            proc = salt.utils.vt.Terminal(
                                cmd,
                                env=env,
                                shell=True,
                                stream_stdout=True,
                                stream_stderr=True,
                            )
                            while proc.has_unread_data:
                                stdout, _ = proc.recv()
                                if stdout and SIGN_PROMPT_RE.search(stdout):
                                    # have the prompt for inputting the passphrase
                                    proc.sendline(phrase)
                                else:
                                    times_looped += 1

                                if times_looped > number_retries:
                                    raise SaltInvocationError(
                                        "Attempting to sign file {} failed, timed out"
                                        " after {} seconds".format(
                                            abs_file, int(times_looped * interval)
                                        )
                                    )
                                time.sleep(interval)

                            proc_exitstatus = proc.exitstatus
                            if proc_exitstatus != 0:
                                raise SaltInvocationError(
                                    "Signing file {} failed with proc.status {}".format(
                                        abs_file, proc_exitstatus
                                    )
                                )
                        except salt.utils.vt.TerminalException as err:
                            trace = traceback.format_exc()
                            log.error(error_msg, err, trace)
                            res = {"retcode": 1, "stdout": "", "stderr": trace}
                        finally:
                            proc.close(terminate=True, kill=True)
                    else:
                        cmd = "debsign --re-sign -k {} {}".format(
                            local_key_fingerprint, abs_file
                        )
                        retrc |= __salt__["cmd.retcode"](
                            cmd, runas=runas, cwd=repodir, use_vt=True, env=env
                        )

                number_retries = timeout / interval
                times_looped = 0
                error_msg = f"Failed to reprepro includedsc file {abs_file}"
                cmd = (
                    "reprepro --ignore=wrongdistribution --component=main -Vb ."
                    " includedsc {} {}".format(codename, abs_file)
                )
                if (
                    __grains__["os"] in ["Ubuntu"] and __grains__["osmajorrelease"] < 18
                ) or (
                    __grains__["os"] in ["Debian"] and __grains__["osmajorrelease"] <= 8
                ):
                    try:
                        proc = salt.utils.vt.Terminal(
                            cmd,
                            env=env,
                            shell=True,
                            cwd=repodir,
                            stream_stdout=True,
                            stream_stderr=True,
                        )
                        while proc.has_unread_data:
                            stdout, _ = proc.recv()
                            if stdout and REPREPRO_SIGN_PROMPT_RE.search(stdout):
                                # have the prompt for inputting the passphrase
                                proc.sendline(phrase)
                            else:
                                times_looped += 1

                            if times_looped > number_retries:
                                raise SaltInvocationError(
                                    "Attempting to reprepro includedsc for file {}"
                                    " failed, timed out after {} loops".format(
                                        abs_file, times_looped
                                    )
                                )
                            time.sleep(interval)

                        proc_exitstatus = proc.exitstatus
                        if proc_exitstatus != 0:
                            raise SaltInvocationError(
                                "Reprepro includedsc for codename {} and file {} failed"
                                " with proc.status {}".format(
                                    codename, abs_file, proc_exitstatus
                                )
                            )
                    except salt.utils.vt.TerminalException as err:
                        trace = traceback.format_exc()
                        log.error(error_msg, err, trace)
                        res = {"retcode": 1, "stdout": "", "stderr": trace}
                    finally:
                        proc.close(terminate=True, kill=True)
                else:
                    retrc |= __salt__["cmd.retcode"](
                        cmd, runas=runas, cwd=repodir, use_vt=True, env=env
                    )

        if retrc != 0:
            raise SaltInvocationError(
                "Making a repo encountered errors, return error {}, check logs for"
                " further details".format(retrc)
            )

        if debfile.endswith(".deb"):
            cmd = (
                "reprepro --ignore=wrongdistribution --component=main -Vb . includedeb"
                " {} {}".format(codename, abs_file)
            )
            res = __salt__["cmd.run_all"](
                cmd, runas=runas, cwd=repodir, use_vt=True, env=env
            )

    return res



"""
Execute batch runs
"""

import copy
import logging
import math
import time
from datetime import datetime, timedelta

import salt.client
import salt.exceptions
import salt.output
import salt.utils.stringutils

log = logging.getLogger(__name__)


class Batch:
    """
    Manage the execution of batch runs

    """

    def __init__(self, opts, eauth=None, quiet=False, _parser=None):
        """
        :param dict opts: A config options dictionary.

        :param dict eauth: An eauth config to use.

                           The default is an empty dict.

        :param bool quiet: Suppress printing to stdout

                           The default is False.
        """
        self.opts = opts
        self.eauth = eauth if eauth else {}
        self.pub_kwargs = eauth if eauth else {}
        self.quiet = quiet
        self.options = _parser
        # Passing listen True to local client will prevent it from purging
        # cahced events while iterating over the batches.
        self.local = salt.client.get_local_client(opts["conf_file"], listen=True)

    def gather_minions(self):
        """
        Return a list of minions to use for the batch run
        """
        args = [
            self.opts["tgt"],
            "test.ping",
            [],
            self.opts["timeout"],
        ]

        selected_target_option = self.opts.get("selected_target_option", None)
        if selected_target_option is not None:
            args.append(selected_target_option)
        else:
            args.append(self.opts.get("tgt_type", "glob"))

        self.pub_kwargs["yield_pub_data"] = True
        ping_gen = self.local.cmd_iter(
            *args, gather_job_timeout=self.opts["gather_job_timeout"], **self.pub_kwargs
        )

        # Broadcast to targets
        fret = set()
        nret = set()
        for ret in ping_gen:
            if ("minions" and "jid") in ret:
                for minion in ret["minions"]:
                    nret.add(minion)
                continue
            else:
                try:
                    m = next(iter(ret.keys()))
                except StopIteration:
                    if not self.quiet:
                        salt.utils.stringutils.print_cli(
                            "No minions matched the target."
                        )
                    break
                if m is not None:
                    if "failed" in ret[m] and ret[m]["failed"] is True:
                        log.debug(
                            "minion '%s' failed test.ping - will be returned as a down minion",
                            m,
                        )
                    else:
                        fret.add(m)

        return (list(fret), ping_gen, nret.difference(fret))

    def get_bnum(self):
        """
        Return the active number of minions to maintain
        """

        def partition(x):
            return float(x) / 100.0 * len(self.minions)

        try:
            if isinstance(self.opts["batch"], str) and "%" in self.opts["batch"]:
                res = partition(float(self.opts["batch"].strip("%")))
                if res < 1:
                    return int(math.ceil(res))
                else:
                    return int(res)
            else:
                return int(self.opts["batch"])
        except ValueError:
            if not self.quiet:
                salt.utils.stringutils.print_cli(
                    "Invalid batch data sent: {}\nData must be in the "
                    "form of %10, 10% or 3".format(self.opts["batch"])
                )

    def __update_wait(self, wait):
        now = datetime.now()
        i = 0
        while i < len(wait) and wait[i] <= now:
            i += 1
        if i:
            del wait[:i]

    def run(self):
        """
        Execute the batch run
        """
        self.minions, self.ping_gen, self.down_minions = self.gather_minions()
        args = [
            [],
            self.opts["fun"],
            self.opts["arg"],
            self.opts["timeout"],
            "list",
        ]
        bnum = self.get_bnum()
        # No targets to run
        if not self.minions:
            return
        to_run = copy.deepcopy(self.minions)
        active = []
        ret = {}
        iters = []
        # wait the specified time before decide a job is actually done
        bwait = self.opts.get("batch_wait", 0)
        wait = []

        if self.options:
            show_jid = self.options.show_jid
            show_verbose = self.options.verbose
        else:
            show_jid = False
            show_verbose = False

        # the minion tracker keeps track of responses and iterators
        # - it removes finished iterators from iters[]
        # - if a previously detected minion does not respond, its
        #   added with an empty answer to ret{} once the timeout is reached
        # - unresponsive minions are removed from active[] to make
        #   sure that the main while loop finishes even with unresp minions
        minion_tracker = {}

        if not self.quiet:
            # We already know some minions didn't respond to the ping, so inform
            # the user we won't be attempting to run a job on them
            for down_minion in self.down_minions:
                salt.utils.stringutils.print_cli(
                    "Minion {} did not respond. No job will be sent.".format(
                        down_minion
                    )
                )

        # Iterate while we still have things to execute
        while len(ret) < len(self.minions):
            next_ = []
            if bwait and wait:
                self.__update_wait(wait)
            if len(to_run) <= bnum - len(wait) and not active:
                # last bit of them, add them all to next iterator
                while to_run:
                    next_.append(to_run.pop())
            else:
                for i in range(bnum - len(active) - len(wait)):
                    if to_run:
                        minion_id = to_run.pop()
                        if isinstance(minion_id, dict):
                            next_.append(next(iter(minion_id)))
                        else:
                            next_.append(minion_id)

            active += next_
            args[0] = next_

            if next_:
                if not self.quiet:
                    salt.utils.stringutils.print_cli(
                        f"\nExecuting run on {sorted(next_)}\n"
                    )
                # create a new iterator for this batch of minions
                return_value = self.opts.get("return", self.opts.get("ret", ""))
                new_iter = self.local.cmd_iter_no_block(
                    *args,
                    raw=self.opts.get("raw", False),
                    ret=return_value,
                    show_jid=show_jid,
                    verbose=show_verbose,
                    gather_job_timeout=self.opts["gather_job_timeout"],
                    **self.eauth,
                )
                # add it to our iterators and to the minion_tracker
                iters.append(new_iter)
                minion_tracker[new_iter] = {}
                # every iterator added is 'active' and has its set of minions
                minion_tracker[new_iter]["minions"] = next_
                minion_tracker[new_iter]["active"] = True

            else:
                time.sleep(0.02)
            parts = {}

            # see if we found more minions
            for ping_ret in self.ping_gen:
                if ping_ret is None:
                    break
                m = next(iter(ping_ret.keys()))
                if m not in self.minions:
                    self.minions.append(m)
                    to_run.append(m)

            for queue in iters:
                try:
                    # Gather returns until we get to the bottom
                    ncnt = 0
                    while True:
                        part = next(queue)
                        if part is None:
                            time.sleep(0.01)
                            ncnt += 1
                            if ncnt > 5:
                                break
                            continue
                        if self.opts.get("raw"):
                            parts.update({part["data"]["id"]: part})
                            if part["data"]["id"] in minion_tracker[queue]["minions"]:
                                minion_tracker[queue]["minions"].remove(
                                    part["data"]["id"]
                                )
                            else:
                                salt.utils.stringutils.print_cli(
                                    "minion {} was already deleted from tracker,"
                                    " probably a duplicate key".format(part["id"])
                                )
                        else:
                            parts.update(part)
                            for id in part:
                                if id in minion_tracker[queue]["minions"]:
                                    minion_tracker[queue]["minions"].remove(id)
                                else:
                                    salt.utils.stringutils.print_cli(
                                        "minion {} was already deleted from tracker,"
                                        " probably a duplicate key".format(id)
                                    )
                except StopIteration:
                    # if a iterator is done:
                    # - set it to inactive
                    # - add minions that have not responded to parts{}

                    # check if the tracker contains the iterator
                    if queue in minion_tracker:
                        minion_tracker[queue]["active"] = False

                        # add all minions that belong to this iterator and
                        # that have not responded to parts{} with an empty response
                        for minion in minion_tracker[queue]["minions"]:
                            if minion not in parts:
                                parts[minion] = {}
                                parts[minion]["ret"] = {}

            for minion, data in parts.items():
                if minion in active:
                    active.remove(minion)
                    if bwait:
                        wait.append(datetime.now() + timedelta(seconds=bwait))
                failhard = False

                # need to check if Minion failed to respond to job sent
                failed_check = data.get("failed", False)
                if failed_check:
                    log.debug(
                        "Minion '%s' failed to respond to job sent, data '%s'",
                        minion,
                        data,
                    )
                    if not self.quiet:
                        # We already know some minions didn't respond to the ping, so inform
                        # inform user attempt to run a job failed
                        salt.utils.stringutils.print_cli(
                            f"Minion '{minion}' failed to respond to job sent"
                        )

                    if self.opts.get("failhard"):
                        failhard = True
                    ret[minion] = data
                else:
                    # If we are executing multiple modules with the same cmd,
                    # We use the highest retcode.
                    retcode = 0
                    if "retcode" in data:
                        if isinstance(data["retcode"], dict):
                            try:
                                data["retcode"] = max(data["retcode"].values())
                            except ValueError:
                                data["retcode"] = 0
                        if self.opts.get("failhard") and data["retcode"] > 0:
                            failhard = True
                        retcode = data["retcode"]

                    if self.opts.get("raw"):
                        ret[minion] = data
                        yield data, retcode
                    else:
                        ret[minion] = data["ret"]
                        yield {minion: data["ret"]}, retcode
                    if not self.quiet:
                        ret[minion] = data["ret"]
                        data[minion] = data.pop("ret")
                        if "out" in data:
                            out = data.pop("out")
                        else:
                            out = None
                        salt.output.display_output(data, out, self.opts)

                if failhard:
                    log.error(
                        "Minion %s returned with non-zero exit code. "
                        "Batch run stopped due to failhard",
                        minion,
                    )
                    return

            # remove inactive iterators from the iters list
            for queue in minion_tracker:
                # only remove inactive queues
                if not minion_tracker[queue]["active"] and queue in iters:
                    iters.remove(queue)
                    # also remove the iterator's minions from the active list
                    for minion in minion_tracker[queue]["minions"]:
                        if minion in active:
                            active.remove(minion)
                            if bwait:
                                wait.append(datetime.now() + timedelta(seconds=bwait))
        self.local.destroy()



"""
    Detect disks
"""

import glob
import logging
import re

# Solve the Chicken and egg problem where grains need to run before any
# of the modules are loaded and are generally available for any usage.
import salt.modules.cmdmod
import salt.utils.files
import salt.utils.path
import salt.utils.platform

__salt__ = {
    "cmd.run": salt.modules.cmdmod._run_quiet,
    "cmd.run_all": salt.modules.cmdmod._run_all_quiet,
    "cmd.powershell": salt.modules.cmdmod.powershell,
}

log = logging.getLogger(__name__)


def disks():
    """
    Return list of disk devices
    """
    if salt.utils.platform.is_freebsd():
        return _freebsd_geom()
    elif salt.utils.platform.is_linux():
        return _linux_disks()
    elif salt.utils.platform.is_windows():
        return _windows_disks()
    else:
        log.trace("Disk grain does not support OS")


class _geomconsts:
    GEOMNAME = "Geom name"
    MEDIASIZE = "Mediasize"
    SECTORSIZE = "Sectorsize"
    STRIPESIZE = "Stripesize"
    STRIPEOFFSET = "Stripeoffset"
    DESCR = "descr"  # model
    LUNID = "lunid"
    LUNNAME = "lunname"
    IDENT = "ident"  # serial
    ROTATIONRATE = "rotationrate"  # RPM or 0 for non-rotating

    # Preserve the API where possible with Salt < 2016.3
    _aliases = {
        DESCR: "device_model",
        IDENT: "serial_number",
        ROTATIONRATE: "media_RPM",
        LUNID: "WWN",
    }

    _datatypes = {
        MEDIASIZE: ("re_int", r"(\d+)"),
        SECTORSIZE: "try_int",
        STRIPESIZE: "try_int",
        STRIPEOFFSET: "try_int",
        ROTATIONRATE: "try_int",
    }


def _datavalue(datatype, data):
    if datatype == "try_int":
        try:
            return int(data)
        except ValueError:
            return None
    elif datatype is tuple and datatype[0] == "re_int":
        search = re.search(datatype[1], data)
        if search:
            try:
                return int(search.group(1))
            except ValueError:
                return None
        return None
    else:
        return data


_geom_attribs = [
    _geomconsts.__dict__[key] for key in _geomconsts.__dict__ if not key.startswith("_")
]


def _freebsd_geom():
    geom = salt.utils.path.which("geom")
    ret = {"disks": {}, "ssds": []}

    devices = __salt__["cmd.run"](f"{geom} disk list")
    devices = devices.split("\n\n")

    def parse_geom_attribs(device):
        tmp = {}
        for line in device.split("\n"):
            for attrib in _geom_attribs:
                search = re.search(rf"{attrib}:\s(.*)", line)
                if search:
                    value = _datavalue(
                        _geomconsts._datatypes.get(attrib), search.group(1)
                    )
                    tmp[attrib] = value
                    if attrib in _geomconsts._aliases:
                        tmp[_geomconsts._aliases[attrib]] = value

        name = tmp.pop(_geomconsts.GEOMNAME)
        if name.startswith("cd"):
            return

        ret["disks"][name] = tmp
        if tmp.get(_geomconsts.ROTATIONRATE) == 0:
            log.trace("Device %s reports itself as an SSD", device)
            ret["ssds"].append(name)

    for device in devices:
        parse_geom_attribs(device)

    return ret


def _linux_disks():
    """
    Return list of disk devices and work out if they are SSD or HDD.
    """
    ret = {"disks": [], "ssds": []}

    for entry in glob.glob("/sys/block/*"):
        virtual = salt.utils.path.readlink(entry).startswith("../devices/virtual/")
        try:
            if not virtual:
                with salt.utils.files.fopen(entry + "/queue/rotational") as entry_fp:
                    device = entry.split("/")[3]
                    flag = entry_fp.read(1)
                    if flag == "0":
                        ret["ssds"].append(device)
                        log.trace("Device %s reports itself as an SSD", device)
                    elif flag == "1":
                        ret["disks"].append(device)
                        log.trace("Device %s reports itself as an HDD", device)
                    else:
                        log.trace(
                            "Unable to identify device %s as an SSD or HDD. It does "
                            "not report 0 or 1",
                            device,
                        )
        except OSError:
            pass
    return ret


def _windows_disks():

    cmd = "Get-PhysicalDisk | Select DeviceID, MediaType"
    ret = {"disks": [], "ssds": []}

    drive_info = __salt__["cmd.powershell"](cmd)

    if not drive_info:
        log.trace("No physical discs found")
        return ret

    # We need a list of dict
    if isinstance(drive_info, dict):
        drive_info = [drive_info]

    for drive in drive_info:
        media_type = drive.get("MediaType")
        # Make sure we have a valid drive type
        if media_type is None or media_type.lower() not in [
            "hdd",
            "ssd",
            "scm",
            "unspecified",
        ]:
            log.trace(f"Unknown media type: {media_type}")
            continue
        device = rf'\\.\PhysicalDrive{drive["DeviceID"]}'
        ret["disks"].append(device)
        if media_type.lower() == "ssd":
            ret["ssds"].append(device)

    return ret



#
# Copyright 2014 SUSE LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Module for managing BTRFS file systems.
"""
import itertools
import os
import re
import subprocess
import uuid

import salt.utils.fsutils
import salt.utils.platform
from salt.exceptions import CommandExecutionError


def __virtual__():
    """
    Only work on POSIX-like systems
    """
    return not salt.utils.platform.is_windows() and __grains__.get("kernel") == "Linux"


def version():
    """
    Return BTRFS version.

    CLI Example:

    .. code-block:: bash

        salt '*' btrfs.version
    """
    out = __salt__["cmd.run_all"]("btrfs --version")
    if out.get("stderr"):
        raise CommandExecutionError(out["stderr"])
    return {"version": out["stdout"].split(" ", 1)[-1]}


def _parse_btrfs_info(data):
    """
    Parse BTRFS device info data.
    """
    ret = {}
    for line in [line for line in data.split("\n") if line][:-1]:
        if line.startswith("Label:"):
            line = re.sub(r"Label:\s+", "", line)
            label, uuid_ = (tkn.strip() for tkn in line.split("uuid:"))
            ret["label"] = label != "none" and label or None
            ret["uuid"] = uuid_
            continue

        if line.startswith("\tdevid"):
            dev_data = re.split(r"\s+", line.strip())
            dev_id = dev_data[-1]
            ret[dev_id] = {
                "device_id": dev_data[1],
                "size": dev_data[3],
                "used": dev_data[5],
            }

    return ret


def info(device):
    """
    Get BTRFS filesystem information.

    CLI Example:

    .. code-block:: bash

        salt '*' btrfs.info /dev/sda1
    """
    out = __salt__["cmd.run_all"](f"btrfs filesystem show {device}")
    salt.utils.fsutils._verify_run(out)

    return _parse_btrfs_info(out["stdout"])


def devices():
    """
    Get known BTRFS formatted devices on the system.

    CLI Example:

    .. code-block:: bash

        salt '*' btrfs.devices
    """
    out = __salt__["cmd.run_all"]("blkid -o export")
    salt.utils.fsutils._verify_run(out)

    return salt.utils.fsutils._blkid_output(out["stdout"], fs_type="btrfs")


def _defragment_mountpoint(mountpoint):
    """
    Defragment only one BTRFS mountpoint.
    """
    out = __salt__["cmd.run_all"](f"btrfs filesystem defragment -f {mountpoint}")
    return {
        "mount_point": mountpoint,
        "passed": not out["stderr"],
        "log": out["stderr"] or False,
        "range": False,
    }


def defragment(path):
    """
    Defragment mounted BTRFS filesystem.
    In order to defragment a filesystem, device should be properly mounted and writable.

    If passed a device name, then defragmented whole filesystem, mounted on in.
    If passed a moun tpoint of the filesystem, then only this mount point is defragmented.

    CLI Example:

    .. code-block:: bash

        salt '*' btrfs.defragment /dev/sda1
        salt '*' btrfs.defragment /path/on/filesystem
    """
    is_device = salt.utils.fsutils._is_device(path)
    mounts = salt.utils.fsutils._get_mounts("btrfs")
    if is_device and not mounts.get(path):
        raise CommandExecutionError(f'Device "{path}" is not mounted')

    result = []
    if is_device:
        for mount_point in mounts[path]:
            result.append(_defragment_mountpoint(mount_point["mount_point"]))
    else:
        is_mountpoint = False
        for mountpoints in mounts.values():
            for mpnt in mountpoints:
                if path == mpnt["mount_point"]:
                    is_mountpoint = True
                    break
        d_res = _defragment_mountpoint(path)
        if (
            not is_mountpoint
            and not d_res["passed"]
            and "range ioctl not supported" in d_res["log"]
        ):
            d_res["log"] = (
                "Range ioctl defragmentation is not supported in this kernel."
            )

        if not is_mountpoint:
            d_res["mount_point"] = False
            d_res["range"] = os.path.exists(path) and path or False

        result.append(d_res)

    return result


def features():
    """
    List currently available BTRFS features.

    CLI Example:

    .. code-block:: bash

        salt '*' btrfs.mkfs_features
    """
    out = __salt__["cmd.run_all"]("mkfs.btrfs -O list-all")
    salt.utils.fsutils._verify_run(out)

    ret = {}
    for line in [
        re.sub(r"\s+", " ", line) for line in out["stderr"].split("\n") if " - " in line
    ]:
        option, description = line.split(" - ", 1)
        ret[option] = description

    return ret


def _usage_overall(raw):
    """
    Parse usage/overall.
    """
    data = {}
    for line in raw.split("\n")[1:]:
        keyset = [
            item.strip()
            for item in re.sub(r"\s+", " ", line).split(":", 1)
            if item.strip()
        ]
        if len(keyset) == 2:
            key = re.sub(r"[()]", "", keyset[0]).replace(" ", "_").lower()
            if key in ["free_estimated", "global_reserve"]:  # An extra field
                subk = keyset[1].split("(")
                data[key] = subk[0].strip()
                subk = subk[1].replace(")", "").split(": ")
                data[f"{key}_{subk[0]}"] = subk[1]
            else:
                data[key] = keyset[1]

    return data


def _usage_specific(raw):
    """
    Parse usage/specific.
    """

    def get_key(val):
        return dict([tuple(val.split(":"))])

    raw = raw.split("\n")
    section, size, used = raw[0].split(" ")
    section = section.replace(",", "_").replace(":", "").lower()

    data = {}
    data[section] = {}

    for val in [size, used]:
        data[section].update(get_key(val.replace(",", "")))

    for devices in raw[1:]:
        data[section].update(get_key(re.sub(r"\s+", ":", devices.strip())))

    return data


def _usage_unallocated(raw):
    """
    Parse usage/unallocated.
    """
    ret = {}
    for line in raw.split("\n")[1:]:
        keyset = re.sub(r"\s+", " ", line.strip()).split(" ")
        if len(keyset) == 2:
            ret[keyset[0]] = keyset[1]

    return ret


def usage(path):
    """
    Show in which disk the chunks are allocated.

    CLI Example:

    .. code-block:: bash

        salt '*' btrfs.usage /your/mountpoint
    """
    out = __salt__["cmd.run_all"](f"btrfs filesystem usage {path}")
    salt.utils.fsutils._verify_run(out)

    ret = {}
    for section in out["stdout"].split("\n\n"):
        if section.startswith("Overall:\n"):
            ret["overall"] = _usage_overall(section)
        elif section.startswith("Unallocated:\n"):
            ret["unallocated"] = _usage_unallocated(section)
        else:
            ret.update(_usage_specific(section))

    return ret


def mkfs(*devices, **kwargs):
    """
    Create a file system on the specified device. By default wipes out with force.

    General options:

    * **allocsize**: Specify the BTRFS offset from the start of the device.
    * **bytecount**: Specify the size of the resultant filesystem.
    * **nodesize**: Node size.
    * **leafsize**: Specify the nodesize, the tree block size in which btrfs stores data.
    * **noforce**: Prevent force overwrite when an existing filesystem is detected on the device.
    * **sectorsize**: Specify the sectorsize, the minimum data block allocation unit.
    * **nodiscard**: Do not perform whole device TRIM operation by default.
    * **uuid**: Pass UUID or pass True to generate one.


    Options:

    * **dto**: (raid0|raid1|raid5|raid6|raid10|single|dup)
               Specify how the data must be spanned across the devices specified.
    * **mto**: (raid0|raid1|raid5|raid6|raid10|single|dup)
               Specify how metadata must be spanned across the devices specified.
    * **fts**: Features (call ``salt <host> btrfs.features`` for full list of available features)

    See the ``mkfs.btrfs(8)`` manpage for a more complete description of corresponding options description.

    CLI Example:

    .. code-block:: bash

        salt '*' btrfs.mkfs /dev/sda1
        salt '*' btrfs.mkfs /dev/sda1 noforce=True
    """
    if not devices:
        raise CommandExecutionError("No devices specified")

    mounts = salt.utils.fsutils._get_mounts("btrfs")
    for device in devices:
        if mounts.get(device):
            raise CommandExecutionError(f'Device "{device}" should not be mounted')

    cmd = ["mkfs.btrfs"]

    dto = kwargs.get("dto")
    mto = kwargs.get("mto")
    if len(devices) == 1:
        if dto:
            cmd.append("-d single")
        if mto:
            cmd.append("-m single")
    else:
        if dto:
            cmd.append(f"-d {dto}")
        if mto:
            cmd.append(f"-m {mto}")

    for key, option in [
        ("-l", "leafsize"),
        ("-L", "label"),
        ("-O", "fts"),
        ("-A", "allocsize"),
        ("-b", "bytecount"),
        ("-n", "nodesize"),
        ("-s", "sectorsize"),
    ]:
        if option == "label" and option in kwargs:
            kwargs["label"] = "'{}'".format(kwargs["label"])
        if kwargs.get(option):
            cmd.append(f"{key} {kwargs.get(option)}")

    if kwargs.get("uuid"):
        cmd.append(
            "-U {}".format(
                kwargs.get("uuid") is True and uuid.uuid1() or kwargs.get("uuid")
            )
        )

    if kwargs.get("nodiscard"):
        cmd.append("-K")
    if not kwargs.get("noforce"):
        cmd.append("-f")

    cmd.extend(devices)

    out = __salt__["cmd.run_all"](" ".join(cmd))
    salt.utils.fsutils._verify_run(out)

    ret = {"log": out["stdout"]}
    ret.update(__salt__["btrfs.info"](devices[0]))

    return ret


def resize(mountpoint, size):
    """
    Resize filesystem.

    General options:

    * **mountpoint**: Specify the BTRFS mountpoint to resize.
    * **size**: ([+/-]<newsize>[kKmMgGtTpPeE]|max) Specify the new size of the target.

    CLI Example:

    .. code-block:: bash

        salt '*' btrfs.resize /mountpoint size=+1g
        salt '*' btrfs.resize /dev/sda1 size=max
    """

    if size == "max":
        if not salt.utils.fsutils._is_device(mountpoint):
            raise CommandExecutionError(
                f'Mountpoint "{mountpoint}" should be a valid device'
            )
        if not salt.utils.fsutils._get_mounts("btrfs").get(mountpoint):
            raise CommandExecutionError(f'Device "{mountpoint}" should be mounted')
    elif (
        len(size) < 3
        or size[0] not in "-+"
        or size[-1] not in "kKmMgGtTpPeE"
        or re.sub(r"\d", "", size[1:][:-1])
    ):
        raise CommandExecutionError(
            'Unknown size: "{}". Expected: [+/-]<newsize>[kKmMgGtTpPeE]|max'.format(
                size
            )
        )

    out = __salt__["cmd.run_all"](f"btrfs filesystem resize {size} {mountpoint}")
    salt.utils.fsutils._verify_run(out)

    ret = {"log": out["stdout"]}
    ret.update(__salt__["btrfs.info"](mountpoint))

    return ret


def _fsck_ext(device):
    """
    Check an ext2/ext3/ext4 file system.

    This is forced check to determine a filesystem is clean or not.
    NOTE: Maybe this function needs to be moved as a standard method in extfs module in a future.
    """
    msgs = {
        0: "No errors",
        1: "Filesystem errors corrected",
        2: "System should be rebooted",
        4: "Filesystem errors left uncorrected",
        8: "Operational error",
        16: "Usage or syntax error",
        32: "Fsck canceled by user request",
        128: "Shared-library error",
    }

    return msgs.get(
        __salt__["cmd.run_all"](f"fsck -f -n {device}")["retcode"],
        "Unknown error",
    )


def convert(device, permanent=False, keeplf=False):
    """
    Convert ext2/3/4 to BTRFS. Device should be mounted.

    Filesystem can be converted temporarily so the further processing and rollback is possible,
    or permanently, where previous extended filesystem image gets deleted. Please note, permanent
    conversion takes a while as BTRFS filesystem needs to be properly rebalanced afterwards.

    General options:

    * **permanent**: Specify if the migration should be permanent (false by default)
    * **keeplf**: Keep ``lost+found`` of the partition (removed by default,
                  but still in the image, if not permanent migration)

    CLI Example:

    .. code-block:: bash

        salt '*' btrfs.convert /dev/sda1
        salt '*' btrfs.convert /dev/sda1 permanent=True
    """

    out = __salt__["cmd.run_all"]("blkid -o export")
    salt.utils.fsutils._verify_run(out)
    devices = salt.utils.fsutils._blkid_output(out["stdout"])
    if not devices.get(device):
        raise CommandExecutionError(f'The device "{device}" was is not found.')

    if not devices[device]["type"] in ["ext2", "ext3", "ext4"]:
        raise CommandExecutionError(
            'The device "{}" is a "{}" file system.'.format(
                device, devices[device]["type"]
            )
        )

    mountpoint = (
        salt.utils.fsutils._get_mounts(devices[device]["type"])
        .get(device, [{"mount_point": None}])[0]
        .get("mount_point")
    )
    if mountpoint == "/":
        raise CommandExecutionError(
            """One does not simply converts a root filesystem!

Converting an extended root filesystem to BTRFS is a careful
and lengthy process, among other steps including the following
requirements:

  1. Proper verified backup.
  2. System outage.
  3. Offline system access.

For further details, please refer to your OS vendor
documentation regarding this topic.
"""
        )

    salt.utils.fsutils._verify_run(__salt__["cmd.run_all"](f"umount {device}"))

    ret = {
        "before": {
            "fsck_status": _fsck_ext(device),
            "mount_point": mountpoint,
            "type": devices[device]["type"],
        }
    }

    salt.utils.fsutils._verify_run(__salt__["cmd.run_all"](f"btrfs-convert {device}"))
    salt.utils.fsutils._verify_run(
        __salt__["cmd.run_all"](f"mount {device} {mountpoint}")
    )

    # Refresh devices
    out = __salt__["cmd.run_all"]("blkid -o export")
    salt.utils.fsutils._verify_run(out)
    devices = salt.utils.fsutils._blkid_output(out["stdout"])

    ret["after"] = {
        "fsck_status": "N/A",  # ToDO
        "mount_point": mountpoint,
        "type": devices[device]["type"],
    }

    # Post-migration procedures
    image_path = f"{mountpoint}/ext2_saved"
    orig_fstype = ret["before"]["type"]

    if not os.path.exists(image_path):
        raise CommandExecutionError(
            f'BTRFS migration went wrong: the image "{image_path}" not found!'
        )

    if not permanent:
        ret["after"][f"{orig_fstype}_image"] = image_path
        image_info_proc = subprocess.run(
            ["file", f"{image_path}/image"], check=True, stdout=subprocess.PIPE
        )
        ret["after"][f"{orig_fstype}_image_info"] = image_info_proc.stdout.strip()
    else:
        ret["after"][f"{orig_fstype}_image"] = "removed"
        ret["after"][f"{orig_fstype}_image_info"] = "N/A"

        salt.utils.fsutils._verify_run(
            __salt__["cmd.run_all"](f"btrfs subvolume delete {image_path}")
        )
        out = __salt__["cmd.run_all"](f"btrfs filesystem balance {mountpoint}")
        salt.utils.fsutils._verify_run(out)
        ret["after"]["balance_log"] = out["stdout"]

    lost_found = f"{mountpoint}/lost+found"
    if os.path.exists(lost_found) and not keeplf:
        salt.utils.fsutils._verify_run(__salt__["cmd.run_all"](f"rm -rf {lost_found}"))

    return ret


def _restripe(mountpoint, direction, *devices, **kwargs):
    """
    Restripe BTRFS: add or remove devices from the particular mounted filesystem.
    """
    fs_log = []

    if salt.utils.fsutils._is_device(mountpoint):
        raise CommandExecutionError(
            f'Mountpount expected, while device "{mountpoint}" specified'
        )

    mounted = False
    for device, mntpoints in salt.utils.fsutils._get_mounts("btrfs").items():
        for mntdata in mntpoints:
            if mntdata["mount_point"] == mountpoint:
                mounted = True
                break

    if not mounted:
        raise CommandExecutionError(
            f'No BTRFS device mounted on "{mountpoint}" mountpoint'
        )

    if not devices:
        raise CommandExecutionError("No devices specified.")

    available_devices = __salt__["btrfs.devices"]()
    for device in devices:
        if device not in available_devices.keys():
            raise CommandExecutionError(f'Device "{device}" is not recognized')

    cmd = [f"btrfs device {direction}"]
    for device in devices:
        cmd.append(device)

    if direction == "add":
        if kwargs.get("nodiscard"):
            cmd.append("-K")
        if kwargs.get("force"):
            cmd.append("-f")

    cmd.append(mountpoint)

    out = __salt__["cmd.run_all"](" ".join(cmd))
    salt.utils.fsutils._verify_run(out)
    if out["stdout"]:
        fs_log.append(out["stdout"])

    if direction == "add":
        out = None
        data_conversion = kwargs.get("dc")
        meta_conversion = kwargs.get("mc")
        if data_conversion and meta_conversion:
            out = __salt__["cmd.run_all"](
                "btrfs balance start -dconvert={} -mconvert={} {}".format(
                    data_conversion, meta_conversion, mountpoint
                )
            )
        else:
            out = __salt__["cmd.run_all"](f"btrfs filesystem balance {mountpoint}")
        salt.utils.fsutils._verify_run(out)
        if out["stdout"]:
            fs_log.append(out["stdout"])

    # Summarize the result
    ret = {}
    if fs_log:
        ret.update({"log": "\n".join(fs_log)})
    ret.update(__salt__["btrfs.info"](mountpoint))

    return ret


def add(mountpoint, *devices, **kwargs):
    """
    Add a devices to a BTRFS filesystem.

    General options:

    * **nodiscard**: Do not perform whole device TRIM
    * **force**: Force overwrite existing filesystem on the disk

    CLI Example:

    .. code-block:: bash

        salt '*' btrfs.add /mountpoint /dev/sda1 /dev/sda2
    """
    return _restripe(mountpoint, "add", *devices, **kwargs)


def delete(mountpoint, *devices, **kwargs):
    """
    Remove devices from a BTRFS filesystem.

    CLI Example:

    .. code-block:: bash

        salt '*' btrfs.delete /mountpoint /dev/sda1 /dev/sda2
    """
    return _restripe(mountpoint, "delete", *devices, **kwargs)


def _parse_proplist(data):
    """
    Parse properties list.
    """
    out = {}
    for line in data.split("\n"):
        line = re.split(r"\s+", line, 1)
        if len(line) == 2:
            out[line[0]] = line[1]

    return out


def properties(obj, type=None, set=None):
    """
    List properties for given btrfs object. The object can be path of BTRFS device,
    mount point, or any directories/files inside the BTRFS filesystem.

    General options:

    * **type**: Possible types are s[ubvol], f[ilesystem], i[node] and d[evice].
    * **force**: Force overwrite existing filesystem on the disk
    * **set**: <key=value,key1=value1...> Options for a filesystem properties.

    CLI Example:

    .. code-block:: bash

        salt '*' btrfs.properties /mountpoint
        salt '*' btrfs.properties /dev/sda1 type=subvol set='ro=false,label="My Storage"'
    """
    if type and type not in [
        "s",
        "subvol",
        "f",
        "filesystem",
        "i",
        "inode",
        "d",
        "device",
    ]:
        raise CommandExecutionError(f'Unknown property type: "{type}" specified')

    cmd = ["btrfs"]
    cmd.append("property")
    cmd.append(set and "set" or "list")
    if type:
        cmd.append(f"-t{type}")
    cmd.append(obj)

    if set:
        try:
            for key, value in [
                [item.strip() for item in keyset.split("=")]
                for keyset in set.split(",")
            ]:
                cmd.append(key)
                cmd.append(value)
        except Exception as ex:  # pylint: disable=broad-except
            raise CommandExecutionError(ex)

    out = __salt__["cmd.run_all"](" ".join(cmd))
    salt.utils.fsutils._verify_run(out)

    if not set:
        ret = {}
        for prop, descr in _parse_proplist(out["stdout"]).items():
            ret[prop] = {"description": descr}
            value = __salt__["cmd.run_all"](f"btrfs property get {obj} {prop}")[
                "stdout"
            ]
            ret[prop]["value"] = value and value.split("=")[-1] or "N/A"

        return ret


def subvolume_exists(path):
    """
    Check if a subvolume is present in the filesystem.

    path
        Mount point for the subvolume (full path)

    CLI Example:

    .. code-block:: bash

        salt '*' btrfs.subvolume_exists /mnt/var

    """
    cmd = ["btrfs", "subvolume", "show", path]
    return __salt__["cmd.retcode"](cmd, ignore_retcode=True) == 0


def subvolume_create(name, dest=None, qgroupids=None):
    """
    Create subvolume `name` in `dest`.

    Return True if the subvolume is created, False is the subvolume is
    already there.

    name
         Name of the new subvolume

    dest
         If not given, the subvolume will be created in the current
         directory, if given will be in /dest/name

    qgroupids
         Add the newly created subcolume to a qgroup. This parameter
         is a list

    CLI Example:

    .. code-block:: bash

        salt '*' btrfs.subvolume_create var
        salt '*' btrfs.subvolume_create var dest=/mnt
        salt '*' btrfs.subvolume_create var qgroupids='[200]'

    """
    if qgroupids and type(qgroupids) is not list:
        raise CommandExecutionError("Qgroupids parameter must be a list")

    if dest:
        name = os.path.join(dest, name)

    # If the subvolume is there, we are done
    if subvolume_exists(name):
        return False

    cmd = ["btrfs", "subvolume", "create"]
    if type(qgroupids) is list:
        cmd.append("-i")
        cmd.extend(qgroupids)
    cmd.append(name)

    res = __salt__["cmd.run_all"](cmd)
    salt.utils.fsutils._verify_run(res)
    return True


def subvolume_delete(name=None, names=None, commit=None):
    """
    Delete the subvolume(s) from the filesystem

    The user can remove one single subvolume (name) or multiple of
    then at the same time (names). One of the two parameters needs to
    specified.

    Please, refer to the documentation to understand the implication
    on the transactions, and when the subvolume is really deleted.

    Return True if the subvolume is deleted, False is the subvolume
    was already missing.

    name
        Name of the subvolume to remove

    names
        List of names of subvolumes to remove

    commit
        * 'after': Wait for transaction commit at the end
        * 'each': Wait for transaction commit after each delete

    CLI Example:

    .. code-block:: bash

        salt '*' btrfs.subvolume_delete /var/volumes/tmp
        salt '*' btrfs.subvolume_delete /var/volumes/tmp commit=after

    """
    if not name and not (names and type(names) is list):
        raise CommandExecutionError("Provide a value for the name parameter")

    if commit and commit not in ("after", "each"):
        raise CommandExecutionError("Value for commit not recognized")

    # Filter the names and take the ones that are still there
    names = [
        n for n in itertools.chain([name], names or []) if n and subvolume_exists(n)
    ]

    # If the subvolumes are gone, we are done
    if not names:
        return False

    cmd = ["btrfs", "subvolume", "delete"]
    if commit == "after":
        cmd.append("--commit-after")
    elif commit == "each":
        cmd.append("--commit-each")
    cmd.extend(names)

    res = __salt__["cmd.run_all"](cmd)
    salt.utils.fsutils._verify_run(res)
    return True


def subvolume_find_new(name, last_gen):
    """
    List the recently modified files in a subvolume

    name
        Name of the subvolume

    last_gen
        Last transid marker from where to compare

    CLI Example:

    .. code-block:: bash

        salt '*' btrfs.subvolume_find_new /var/volumes/tmp 1024

    """
    cmd = ["btrfs", "subvolume", "find-new", name, last_gen]

    res = __salt__["cmd.run_all"](cmd)
    salt.utils.fsutils._verify_run(res)

    lines = res["stdout"].splitlines()
    # Filenames are at the end of each inode line
    files = [l.split()[-1] for l in lines if l.startswith("inode")]
    # The last transid is in the last line
    transid = lines[-1].split()[-1]
    return {
        "files": files,
        "transid": transid,
    }


def subvolume_get_default(path):
    """
    Get the default subvolume of the filesystem path

    path
        Mount point for the subvolume

    CLI Example:

    .. code-block:: bash

        salt '*' btrfs.subvolume_get_default /var/volumes/tmp

    """
    cmd = ["btrfs", "subvolume", "get-default", path]

    res = __salt__["cmd.run_all"](cmd)
    salt.utils.fsutils._verify_run(res)

    line = res["stdout"].strip()
    # The ID is the second parameter, and the name the last one, or
    # '(FS_TREE)'
    #
    # When the default one is set:
    # ID 5 (FS_TREE)
    #
    # When we manually set a different one (var):
    # ID 257 gen 8 top level 5 path var
    #
    id_ = line.split()[1]
    name = line.split()[-1]
    return {
        "id": id_,
        "name": name,
    }


def _pop(line, key, use_rest):
    """
    Helper for the line parser.

    If key is a prefix of line, will remove ir from the line and will
    extract the value (space separation), and the rest of the line.

    If use_rest is True, the value will be the rest of the line.

    Return a tuple with the value and the rest of the line.
    """
    value = None
    if line.startswith(key):
        line = line[len(key) :].strip()
        if use_rest:
            value = line
            line = ""
        else:
            value, line = line.split(" ", 1)
    return value, line.strip()


def subvolume_list(
    path,
    parent_id=False,
    absolute=False,
    ogeneration=False,
    generation=False,
    subvolumes=False,
    uuid=False,
    parent_uuid=False,
    sent_subvolume_uuid=False,
    snapshots=False,
    readonly=False,
    deleted=False,
    generation_cmp=None,
    ogeneration_cmp=None,
    sort=None,
):
    """
    List the subvolumes present in the filesystem.

    path
        Mount point for the subvolume

    parent_id
        Print parent ID

    absolute
        Print all the subvolumes in the filesystem and distinguish
        between absolute and relative path with respect to the given
        <path>

    ogeneration
        Print the ogeneration of the subvolume

    generation
        Print the generation of the subvolume

    subvolumes
        Print only subvolumes below specified <path>

    uuid
        Print the UUID of the subvolume

    parent_uuid
        Print the parent uuid of subvolumes (and snapshots)

    sent_subvolume_uuid
        Print the UUID of the sent subvolume, where the subvolume is
        the result of a receive operation

    snapshots
        Only snapshot subvolumes in the filesystem will be listed

    readonly
        Only readonly subvolumes in the filesystem will be listed

    deleted
        Only deleted subvolumens that are ye not cleaned

    generation_cmp
        List subvolumes in the filesystem that its generation is >=,
        <= or = value. '+' means >= value, '-' means <= value, If
        there is neither '+' nor '-', it means = value

    ogeneration_cmp
        List subvolumes in the filesystem that its ogeneration is >=,
        <= or = value

    sort
        List subvolumes in order by specified items. Possible values:
        * rootid
        * gen
        * ogen
        * path
        You can add '+' or '-' in front of each items, '+' means
        ascending, '-' means descending. The default is ascending. You
        can combite it in a list.

    CLI Example:

    .. code-block:: bash

        salt '*' btrfs.subvolume_list /var/volumes/tmp
        salt '*' btrfs.subvolume_list /var/volumes/tmp path=True
        salt '*' btrfs.subvolume_list /var/volumes/tmp sort='[-rootid]'

    """
    if sort and type(sort) is not list:
        raise CommandExecutionError("Sort parameter must be a list")

    valid_sorts = [
        "".join((order, attrib))
        for order, attrib in itertools.product(
            ("-", "", "+"), ("rootid", "gen", "ogen", "path")
        )
    ]
    if sort and not all(s in valid_sorts for s in sort):
        raise CommandExecutionError("Value for sort not recognized")

    cmd = ["btrfs", "subvolume", "list"]

    params = (
        (parent_id, "-p"),
        (absolute, "-a"),
        (ogeneration, "-c"),
        (generation, "-g"),
        (subvolumes, "-o"),
        (uuid, "-u"),
        (parent_uuid, "-q"),
        (sent_subvolume_uuid, "-R"),
        (snapshots, "-s"),
        (readonly, "-r"),
        (deleted, "-d"),
    )
    cmd.extend(p[1] for p in params if p[0])

    if generation_cmp:
        cmd.extend(["-G", generation_cmp])

    if ogeneration_cmp:
        cmd.extend(["-C", ogeneration_cmp])

    # We already validated the content of the list
    if sort:
        cmd.append("--sort={}".format(",".join(sort)))

    cmd.append(path)

    res = __salt__["cmd.run_all"](cmd)
    salt.utils.fsutils._verify_run(res)

    # Parse the output. ID and gen are always at the beginning, and
    # path is always at the end. There is only one column that
    # contains space (top level), and the path value can also have
    # spaces. The issue is that we do not know how many spaces do we
    # have in the path name, so any classic solution based on split
    # will fail.
    #
    # This list is in order.
    columns = (
        "ID",
        "gen",
        "cgen",
        "parent",
        "top level",
        "otime",
        "parent_uuid",
        "received_uuid",
        "uuid",
        "path",
    )
    result = []
    for line in res["stdout"].splitlines():
        table = {}
        for key in columns:
            value, line = _pop(line, key, key == "path")
            if value:
                table[key.lower()] = value
        # If line is not empty here, we are not able to parse it
        if not line:
            result.append(table)

    return result


def subvolume_set_default(subvolid, path):
    """
    Set the subvolume as default

    subvolid
        ID of the new default subvolume

    path
        Mount point for the filesystem

    CLI Example:

    .. code-block:: bash

        salt '*' btrfs.subvolume_set_default 257 /var/volumes/tmp

    """
    cmd = ["btrfs", "subvolume", "set-default", subvolid, path]

    res = __salt__["cmd.run_all"](cmd)
    salt.utils.fsutils._verify_run(res)
    return True


def subvolume_show(path):
    """
    Show information of a given subvolume

    path
        Mount point for the filesystem

    CLI Example:

    .. code-block:: bash

        salt '*' btrfs.subvolume_show /var/volumes/tmp

    """
    cmd = ["btrfs", "subvolume", "show", path]

    res = __salt__["cmd.run_all"](cmd)
    salt.utils.fsutils._verify_run(res)

    result = {}
    table = {}
    # The real name is the first line, later there is a table of
    # values separated with colon.
    stdout = res["stdout"].splitlines()
    key = stdout.pop(0)
    result[key.strip()] = table

    for line in stdout:
        key, value = line.split(":", 1)
        table[key.lower().strip()] = value.strip()
    return result


def subvolume_snapshot(source, dest=None, name=None, read_only=False):
    """
    Create a snapshot of a source subvolume

    source
        Source subvolume from where to create the snapshot

    dest
        If only dest is given, the subvolume will be named as the
        basename of the source

    name
       Name of the snapshot

    read_only
        Create a read only snapshot

    CLI Example:

    .. code-block:: bash

        salt '*' btrfs.subvolume_snapshot /var/volumes/tmp dest=/.snapshots
        salt '*' btrfs.subvolume_snapshot /var/volumes/tmp name=backup

    """
    if not dest and not name:
        raise CommandExecutionError("Provide parameter dest, name, or both")

    cmd = ["btrfs", "subvolume", "snapshot"]
    if read_only:
        cmd.append("-r")

    cmd.append(source)

    if dest and not name:
        cmd.append(dest)
    if dest and name:
        name = os.path.join(dest, name)
    if name:
        cmd.append(name)

    res = __salt__["cmd.run_all"](cmd)
    salt.utils.fsutils._verify_run(res)
    return True


def subvolume_sync(path, subvolids=None, sleep=None):
    """
    Wait until given subvolume are completely removed from the
    filesystem after deletion.

    path
        Mount point for the filesystem

    subvolids
        List of IDs of subvolumes to wait for

    sleep
        Sleep N seconds betwenn checks (default: 1)

    CLI Example:

    .. code-block:: bash

        salt '*' btrfs.subvolume_sync /var/volumes/tmp
        salt '*' btrfs.subvolume_sync /var/volumes/tmp subvolids='[257]'

    """
    if subvolids and type(subvolids) is not list:
        raise CommandExecutionError("Subvolids parameter must be a list")

    cmd = ["btrfs", "subvolume", "sync"]
    if sleep:
        cmd.extend(["-s", sleep])

    cmd.append(path)
    if subvolids:
        cmd.extend(subvolids)

    res = __salt__["cmd.run_all"](cmd)
    salt.utils.fsutils._verify_run(res)
    return True



"""
Management of SmartOS Standalone Compute Nodes

:maintainer:    Jorge Schrauwen <sjorge@blackdot.be>
:maturity:      new
:depends:       vmadm, imgadm
:platform:      smartos

.. versionadded:: 2016.3.0

.. code-block:: yaml

    vmtest.example.org:
      smartos.vm_present:
        - config:
            reprovision: true
        - vmconfig:
            image_uuid: c02a2044-c1bd-11e4-bd8c-dfc1db8b0182
            brand: joyent
            alias: vmtest
            quota: 5
            max_physical_memory: 512
            tags:
              label: 'test vm'
              owner: 'sjorge'
            nics:
              "82:1b:8e:49:e9:12":
                nic_tag: trunk
                mtu: 1500
                ips:
                  - 172.16.1.123/16
                  - 192.168.2.123/24
                vlan_id: 10
              "82:1b:8e:49:e9:13":
                nic_tag: trunk
                mtu: 1500
                ips:
                  - dhcp
                vlan_id: 30
            filesystems:
              "/bigdata":
                source: "/bulk/data"
                type: lofs
                options:
                  - ro
                  - nodevices

    kvmtest.example.org:
      smartos.vm_present:
        - vmconfig:
            brand: kvm
            alias: kvmtest
            cpu_type: host
            ram: 512
            vnc_port: 9
            tags:
              label: 'test kvm'
              owner: 'sjorge'
            disks:
              disk0:
                size: 2048
                model: virtio
                compression: lz4
                boot: true
            nics:
              "82:1b:8e:49:e9:15":
                nic_tag: trunk
                mtu: 1500
                ips:
                  - dhcp
                vlan_id: 30

    docker.example.org:
      smartos.vm_present:
        - config:
            auto_import: true
            reprovision: true
        - vmconfig:
            image_uuid: emby/embyserver:latest
            brand: lx
            alias: mydockervm
            quota: 5
            max_physical_memory: 1024
            tags:
              label: 'my emby docker'
              owner: 'sjorge'
            resolvers:
              - 172.16.1.1
            nics:
              "82:1b:8e:49:e9:18":
                nic_tag: trunk
                mtu: 1500
                ips:
                  - 172.16.1.118/24
                vlan_id: 10
            filesystems:
              "/config:
                source: "/vmdata/emby_config"
                type: lofs
                options:
                  - nodevices

    cleanup_images:
      smartos.image_vacuum

.. note::

    Keep in mind that when removing properties from vmconfig they will not get
    removed from the vm's current configuration, except for nics, disk, tags, ...
    they get removed via add_*, set_*, update_*, and remove_*. Properties must
    be manually reset to their default value.
    The same behavior as when using 'vmadm update'.

.. warning::

    For HVM (bhyve and KVM) brands the `image_uuid` field should go on the boot disks,
    this disk should NOT have a size specified. (See man vmadm)

"""

import json
import logging
import os

import salt.utils.atomicfile
import salt.utils.data
import salt.utils.files

log = logging.getLogger(__name__)

# Define the state's virtual name
__virtualname__ = "smartos"


def __virtual__():
    """
    Provides smartos state provided for SmartOS
    """
    if "vmadm.create" in __salt__ and "imgadm.list" in __salt__:
        return True
    else:
        return (
            False,
            "{} state module can only be loaded on SmartOS compute nodes".format(
                __virtualname__
            ),
        )


def _split_docker_uuid(uuid):
    """
    Split a smartos docker uuid into repo and tag
    """
    if uuid:
        uuid = uuid.split(":")
        if len(uuid) == 2:
            tag = uuid[1]
            repo = uuid[0]
            return repo, tag
    return None, None


def _is_uuid(uuid):
    """
    Check if uuid is a valid smartos uuid

    Example: e69a0918-055d-11e5-8912-e3ceb6df4cf8
    """
    if uuid and list(len(x) for x in uuid.split("-")) == [8, 4, 4, 4, 12]:
        return True
    return False


def _is_docker_uuid(uuid):
    """
    Check if uuid is a valid smartos docker uuid

    Example plexinc/pms-docker:plexpass
    """
    repo, tag = _split_docker_uuid(uuid)
    return not (not repo and not tag)


def _load_config():
    """
    Loads and parses /usbkey/config
    """
    config = {}

    if os.path.isfile("/usbkey/config"):
        with salt.utils.files.fopen("/usbkey/config", "r") as config_file:
            for optval in config_file:
                optval = salt.utils.stringutils.to_unicode(optval)
                if optval[0] == "#":
                    continue
                if "=" not in optval:
                    continue
                optval = optval.split("=")
                config[optval[0].lower()] = optval[1].strip().strip('"')
    log.debug("smartos.config - read /usbkey/config: %s", config)
    return config


def _write_config(config):
    """
    writes /usbkey/config
    """
    try:
        with salt.utils.atomicfile.atomic_open("/usbkey/config", "w") as config_file:
            config_file.write("#\n# This file was generated by salt\n#\n")
            for prop in salt.utils.odict.OrderedDict(sorted(config.items())):
                if " " in str(config[prop]):
                    if not config[prop].startswith('"') or not config[prop].endswith(
                        '"'
                    ):
                        config[prop] = f'"{config[prop]}"'
                config_file.write(
                    salt.utils.stringutils.to_str(f"{prop}={config[prop]}\n")
                )
        log.debug("smartos.config - wrote /usbkey/config: %s", config)
    except OSError:
        return False

    return True


def _parse_vmconfig(config, instances):
    """
    Parse vm_present vm config
    """
    vmconfig = None

    if isinstance(config, (salt.utils.odict.OrderedDict)):
        vmconfig = salt.utils.odict.OrderedDict()
        for prop in config:
            if prop not in instances:
                vmconfig[prop] = config[prop]
            else:
                if not isinstance(config[prop], (salt.utils.odict.OrderedDict)):
                    continue
                vmconfig[prop] = []
                for instance in config[prop]:
                    instance_config = config[prop][instance]
                    instance_config[instances[prop]] = instance
                    ## some property are lowercase
                    if "mac" in instance_config:
                        instance_config["mac"] = instance_config["mac"].lower()
                    ## calculate mac from vrrp_vrid
                    if "vrrp_vrid" in instance_config:
                        instance_config["mac"] = "00:00:5e:00:01:{}".format(
                            hex(int(instance_config["vrrp_vrid"]))
                            .split("x")[-1]
                            .zfill(2),
                        )
                    vmconfig[prop].append(instance_config)
    else:
        log.error("smartos.vm_present::parse_vmconfig - failed to parse")

    return vmconfig


def _get_instance_changes(current, state):
    """
    get modified properties
    """
    # get keys
    current_keys = set(current.keys())
    state_keys = set(state.keys())

    # compare configs
    changed = salt.utils.data.compare_dicts(current, state)
    for change in salt.utils.data.compare_dicts(current, state):
        if change in changed and changed[change]["old"] == "":
            del changed[change]
        if change in changed and changed[change]["new"] == "":
            del changed[change]

    return changed


def _copy_lx_vars(vmconfig):
    # NOTE: documentation on dockerinit: https://github.com/joyent/smartos-live/blob/master/src/dockerinit/README.md
    if "image_uuid" in vmconfig:
        # NOTE: retrieve tags and type from image
        imgconfig = __salt__["imgadm.get"](vmconfig["image_uuid"]).get("manifest", {})
        imgtype = imgconfig.get("type", "zone-dataset")
        imgtags = imgconfig.get("tags", {})

        # NOTE: copy kernel_version (if not specified in vmconfig)
        if "kernel_version" not in vmconfig and "kernel_version" in imgtags:
            vmconfig["kernel_version"] = imgtags["kernel_version"]

        # NOTE: copy docker vars
        if imgtype == "docker":
            vmconfig["docker"] = True
            vmconfig["kernel_version"] = vmconfig.get("kernel_version", "4.3.0")
            if "internal_metadata" not in vmconfig:
                vmconfig["internal_metadata"] = {}

            for var in imgtags.get("docker:config", {}):
                val = imgtags["docker:config"][var]
                var = f"docker:{var.lower()}"

                # NOTE: skip empty values
                if not val:
                    continue

                # NOTE: skip or merge user values
                if var == "docker:env":
                    try:
                        val_config = json.loads(
                            vmconfig["internal_metadata"].get(var, "")
                        )
                    except ValueError as e:
                        val_config = []

                    for config_env_var in (
                        val_config
                        if isinstance(val_config, list)
                        else json.loads(val_config)
                    ):
                        config_env_var = config_env_var.split("=")
                        for img_env_var in val:
                            if img_env_var.startswith(f"{config_env_var[0]}="):
                                val.remove(img_env_var)
                        val.append("=".join(config_env_var))
                elif var in vmconfig["internal_metadata"]:
                    continue

                if isinstance(val, list):
                    # NOTE: string-encoded JSON arrays
                    vmconfig["internal_metadata"][var] = json.dumps(val)
                else:
                    vmconfig["internal_metadata"][var] = val

    return vmconfig


def config_present(name, value):
    """
    Ensure configuration property is set to value in /usbkey/config

    name : string
        name of property
    value : string
        value of property

    """
    name = name.lower()
    ret = {"name": name, "changes": {}, "result": None, "comment": ""}

    # load confiration
    config = _load_config()

    # handle bool and None value
    if isinstance(value, (bool)):
        value = "true" if value else "false"
    if not value:
        value = ""

    if name in config:
        if str(config[name]) == str(value):
            # we're good
            ret["result"] = True
            ret["comment"] = f'property {name} already has value "{value}"'
        else:
            # update property
            ret["result"] = True
            ret["comment"] = f'updated property {name} with value "{value}"'
            ret["changes"][name] = value
            config[name] = value
    else:
        # add property
        ret["result"] = True
        ret["comment"] = f'added property {name} with value "{value}"'
        ret["changes"][name] = value
        config[name] = value

    # apply change if needed
    if not __opts__["test"] and ret["changes"]:
        ret["result"] = _write_config(config)

        if not ret["result"]:
            ret["comment"] = (
                'Could not add property {} with value "{}" to config'.format(
                    name, value
                )
            )

    return ret


def config_absent(name):
    """
    Ensure configuration property is absent in /usbkey/config

    name : string
        name of property

    """
    name = name.lower()
    ret = {"name": name, "changes": {}, "result": None, "comment": ""}

    # load configuration
    config = _load_config()

    if name in config:
        # delete property
        ret["result"] = True
        ret["comment"] = f"property {name} deleted"
        ret["changes"][name] = None
        del config[name]
    else:
        # we're good
        ret["result"] = True
        ret["comment"] = f"property {name} is absent"

    # apply change if needed
    if not __opts__["test"] and ret["changes"]:
        ret["result"] = _write_config(config)

    return ret


def source_present(name, source_type="imgapi"):
    """
    Ensure an image source is present on the computenode

    name : string
        source url
    source_type : string
        source type (imgapi or docker)
    """
    ret = {"name": name, "changes": {}, "result": None, "comment": ""}

    if name in __salt__["imgadm.sources"]():
        # source is present
        ret["result"] = True
        ret["comment"] = f"image source {name} is present"
    else:
        # add new source
        if __opts__["test"]:
            res = {}
            ret["result"] = True
        else:
            res = __salt__["imgadm.source_add"](name, source_type)
            ret["result"] = name in res

        if ret["result"]:
            ret["comment"] = f"image source {name} added"
            ret["changes"][name] = "added"
        else:
            ret["comment"] = f"image source {name} not added"
            if "Error" in res:
                ret["comment"] = "{}: {}".format(ret["comment"], res["Error"])

    return ret


def source_absent(name):
    """
    Ensure an image source is absent on the computenode

    name : string
        source url
    """
    ret = {"name": name, "changes": {}, "result": None, "comment": ""}

    if name not in __salt__["imgadm.sources"]():
        # source is absent
        ret["result"] = True
        ret["comment"] = f"image source {name} is absent"
    else:
        # remove source
        if __opts__["test"]:
            res = {}
            ret["result"] = True
        else:
            res = __salt__["imgadm.source_delete"](name)
            ret["result"] = name not in res

        if ret["result"]:
            ret["comment"] = f"image source {name} deleted"
            ret["changes"][name] = "deleted"
        else:
            ret["comment"] = f"image source {name} not deleted"
            if "Error" in res:
                ret["comment"] = "{}: {}".format(ret["comment"], res["Error"])

    return ret


def image_present(name):
    """
    Ensure image is present on the computenode

    name : string
        uuid of image
    """
    ret = {"name": name, "changes": {}, "result": None, "comment": ""}

    if _is_docker_uuid(name) and __salt__["imgadm.docker_to_uuid"](name):
        # docker image was imported
        ret["result"] = True
        ret["comment"] = "image {} ({}) is present".format(
            name,
            __salt__["imgadm.docker_to_uuid"](name),
        )
    elif name in __salt__["imgadm.list"]():
        # image was already imported
        ret["result"] = True
        ret["comment"] = f"image {name} is present"
    else:
        # add image
        if _is_docker_uuid(name):
            # NOTE: we cannot query available docker images
            available_images = [name]
        else:
            available_images = __salt__["imgadm.avail"]()

        if name in available_images:
            if __opts__["test"]:
                ret["result"] = True
                res = {}
                if _is_docker_uuid(name):
                    res["00000000-0000-0000-0000-000000000000"] = name
                else:
                    res[name] = available_images[name]
            else:
                res = __salt__["imgadm.import"](name)
                if _is_uuid(name):
                    ret["result"] = name in res
                elif _is_docker_uuid(name):
                    ret["result"] = __salt__["imgadm.docker_to_uuid"](name) is not None
            if ret["result"]:
                ret["comment"] = f"image {name} imported"
                ret["changes"] = res
            else:
                ret["comment"] = f"image {name} was unable to be imported"
        else:
            ret["result"] = False
            ret["comment"] = f"image {name} does not exists"

    return ret


def image_absent(name):
    """
    Ensure image is absent on the computenode

    name : string
        uuid of image

    .. note::

        computenode.image_absent will only remove the image if it is not used
        by a vm.
    """
    ret = {"name": name, "changes": {}, "result": None, "comment": ""}

    uuid = None
    if _is_uuid(name):
        uuid = name
    if _is_docker_uuid(name):
        uuid = __salt__["imgadm.docker_to_uuid"](name)

    if not uuid or uuid not in __salt__["imgadm.list"]():
        # image not imported
        ret["result"] = True
        ret["comment"] = f"image {name} is absent"
    else:
        # check if image in use by vm
        if uuid in __salt__["vmadm.list"](order="image_uuid"):
            ret["result"] = False
            ret["comment"] = f"image {name} currently in use by a vm"
        else:
            # delete image
            if __opts__["test"]:
                ret["result"] = True
            else:
                image = __salt__["imgadm.get"](uuid)
                image_count = 0
                if image["manifest"]["name"] == "docker-layer":
                    # NOTE: docker images are made of multiple layers, loop over them
                    while image:
                        image_count += 1
                        __salt__["imgadm.delete"](image["manifest"]["uuid"])
                        if "origin" in image["manifest"]:
                            image = __salt__["imgadm.get"](image["manifest"]["origin"])
                        else:
                            image = None
                else:
                    # NOTE: normal images can just be delete
                    __salt__["imgadm.delete"](uuid)

            ret["result"] = uuid not in __salt__["imgadm.list"]()
            if image_count:
                ret["comment"] = "image {} and {} children deleted".format(
                    name, image_count
                )
            else:
                ret["comment"] = f"image {name} deleted"
            ret["changes"][name] = None

    return ret


def image_vacuum(name):
    """
    Delete images not in use or installed via image_present

    .. warning::

        Only image_present states that are included via the
        top file will be detected.
    """
    name = name.lower()
    ret = {"name": name, "changes": {}, "result": None, "comment": ""}

    # list of images to keep
    images = []

    # retrieve image_present state data for host
    for state in __salt__["state.show_lowstate"]():
        # don't throw exceptions when not highstate run
        if "state" not in state:
            continue

        # skip if not from this state module
        if state["state"] != __virtualname__:
            continue
        # skip if not image_present
        if state["fun"] not in ["image_present"]:
            continue
        # keep images installed via image_present
        if "name" in state:
            if _is_uuid(state["name"]):
                images.append(state["name"])
            elif _is_docker_uuid(state["name"]):
                state["name"] = __salt__["imgadm.docker_to_uuid"](state["name"])
                if not state["name"]:
                    continue
                images.append(state["name"])

    # retrieve images in use by vms
    for image_uuid in __salt__["vmadm.list"](order="image_uuid"):
        if image_uuid not in images:
            images.append(image_uuid)

    # purge unused images
    ret["result"] = True
    for image_uuid in __salt__["imgadm.list"]():
        if image_uuid in images:
            continue

        image = __salt__["imgadm.get"](image_uuid)
        if image["manifest"]["name"] == "docker-layer":
            # NOTE: docker images are made of multiple layers, loop over them
            while image:
                image_uuid = image["manifest"]["uuid"]
                if image_uuid in __salt__["imgadm.delete"](image_uuid):
                    ret["changes"][image_uuid] = None
                else:
                    ret["result"] = False
                    ret["comment"] = "failed to delete images"
                if "origin" in image["manifest"]:
                    image = __salt__["imgadm.get"](image["manifest"]["origin"])
                else:
                    image = None
        else:
            # NOTE: normal images can just be delete
            if image_uuid in __salt__["imgadm.delete"](image_uuid):
                ret["changes"][image_uuid] = None
            else:
                ret["result"] = False
                ret["comment"] = "failed to delete images"

    if ret["result"] and not ret["changes"]:
        ret["comment"] = "no images deleted"
    elif ret["result"] and ret["changes"]:
        ret["comment"] = "images deleted"

    return ret


def vm_present(name, vmconfig, config=None):
    """
    Ensure vm is present on the computenode

    name : string
        hostname of vm
    vmconfig : dict
        options to set for the vm
    config : dict
        fine grain control over vm_present

    .. note::

        The following configuration properties can be toggled in the config parameter.
          - kvm_reboot (true)                - reboots of kvm zones if needed for a config update
          - auto_import (false)              - automatic importing of missing images
          - auto_lx_vars (true)              - copy kernel_version and docker:* variables from image
          - reprovision (false)              - reprovision on image_uuid changes
          - enforce_tags (true)              - false = add tags only, true =  add, update, and remove tags
          - enforce_routes (true)            - false = add tags only, true =  add, update, and remove routes
          - enforce_internal_metadata (true) - false = add metadata only, true =  add, update, and remove metadata
          - enforce_customer_metadata (true) - false = add metadata only, true =  add, update, and remove metadata

    .. note::

        State ID is used as hostname. Hostnames must be unique.

    .. note::

        If hostname is provided in vmconfig this will take president over the State ID.
        This allows multiple states to be applied to the same vm.

    .. note::

        The following instances should have a unique ID.
          - nic : mac
          - filesystem: target
          - disk : path or diskN for zvols

        e.g. disk0 will be the first disk added, disk1 the 2nd,...

    .. versionchanged:: 2019.2.0

        Added support for docker image uuids, added auto_lx_vars configuration, documented some missing configuration options.

    """
    name = name.lower()
    ret = {"name": name, "changes": {}, "result": None, "comment": ""}

    # config defaults
    state_config = config if config else {}
    config = {
        "kvm_reboot": True,
        "auto_import": False,
        "auto_lx_vars": True,
        "reprovision": False,
        "enforce_tags": True,
        "enforce_routes": True,
        "enforce_internal_metadata": True,
        "enforce_customer_metadata": True,
    }
    config.update(state_config)
    log.debug("smartos.vm_present::%s::config - %s", name, config)

    # map special vmconfig parameters
    #  collections have set/remove handlers
    #  instances have add/update/remove handlers and a unique id
    vmconfig_type = {
        "collection": ["tags", "customer_metadata", "internal_metadata", "routes"],
        "instance": {
            "nics": "mac",
            "disks": "path",
            "filesystems": "target",
            "pci_devices": "path",
        },
        "create_only": ["filesystems"],
    }
    vmconfig_docker_keep = [
        "docker:id",
        "docker:restartcount",
    ]
    vmconfig_docker_array = [
        "docker:env",
        "docker:cmd",
        "docker:entrypoint",
    ]

    # parse vmconfig
    vmconfig = _parse_vmconfig(vmconfig, vmconfig_type["instance"])
    log.debug("smartos.vm_present::%s::vmconfig - %s", name, vmconfig)

    # set hostname if needed
    if "hostname" not in vmconfig:
        vmconfig["hostname"] = name

    # prepare image_uuid
    if "image_uuid" in vmconfig:
        # NOTE: lookup uuid from docker uuid (normal uuid's are passed throuhg unmodified)
        #       we must do this again if we end up importing a missing image later!
        docker_uuid = __salt__["imgadm.docker_to_uuid"](vmconfig["image_uuid"])
        vmconfig["image_uuid"] = docker_uuid if docker_uuid else vmconfig["image_uuid"]

        # NOTE: import image (if missing and allowed)
        if vmconfig["image_uuid"] not in __salt__["imgadm.list"]():
            if config["auto_import"]:
                if not __opts__["test"]:
                    res = __salt__["imgadm.import"](vmconfig["image_uuid"])
                    vmconfig["image_uuid"] = __salt__["imgadm.docker_to_uuid"](
                        vmconfig["image_uuid"]
                    )
                    if vmconfig["image_uuid"] not in res:
                        ret["result"] = False
                        ret["comment"] = "failed to import image {}".format(
                            vmconfig["image_uuid"]
                        )
            else:
                ret["result"] = False
                ret["comment"] = "image {} not installed".format(vmconfig["image_uuid"])

    # prepare disk.*.image_uuid
    for disk in vmconfig["disks"] if "disks" in vmconfig else []:
        if "image_uuid" in disk and disk["image_uuid"] not in __salt__["imgadm.list"]():
            if config["auto_import"]:
                if not __opts__["test"]:
                    res = __salt__["imgadm.import"](disk["image_uuid"])
                    if disk["image_uuid"] not in res:
                        ret["result"] = False
                        ret["comment"] = "failed to import image {}".format(
                            disk["image_uuid"]
                        )
            else:
                ret["result"] = False
                ret["comment"] = "image {} not installed".format(disk["image_uuid"])

    # docker json-array handling
    if "internal_metadata" in vmconfig:
        for var in vmconfig_docker_array:
            if var not in vmconfig["internal_metadata"]:
                continue
            if isinstance(vmconfig["internal_metadata"][var], list):
                vmconfig["internal_metadata"][var] = json.dumps(
                    vmconfig["internal_metadata"][var]
                )

    # copy lx variables
    if vmconfig["brand"] == "lx" and config["auto_lx_vars"]:
        # NOTE: we can only copy the lx vars after the image has bene imported
        vmconfig = _copy_lx_vars(vmconfig)

    # quick abort if things look wrong
    # NOTE: use explicit check for false, otherwise None also matches!
    if ret["result"] is False:
        return ret

    # check if vm exists
    if vmconfig["hostname"] in __salt__["vmadm.list"](order="hostname"):
        # update vm
        ret["result"] = True

        # expand vmconfig
        vmconfig = {
            "state": vmconfig,
            "current": __salt__["vmadm.get"](vmconfig["hostname"], key="hostname"),
            "changed": {},
            "reprovision_uuid": None,
        }

        # prepare reprovision
        if "image_uuid" in vmconfig["state"]:
            vmconfig["reprovision_uuid"] = vmconfig["state"]["image_uuid"]
            vmconfig["state"]["image_uuid"] = vmconfig["current"]["image_uuid"]

        # disks need some special care
        if "disks" in vmconfig["state"]:
            new_disks = []
            for disk in vmconfig["state"]["disks"]:
                path = False
                if "disks" in vmconfig["current"]:
                    for cdisk in vmconfig["current"]["disks"]:
                        if cdisk["path"].endswith(disk["path"]):
                            path = cdisk["path"]
                            break
                if not path:
                    del disk["path"]
                else:
                    disk["path"] = path
                new_disks.append(disk)
            vmconfig["state"]["disks"] = new_disks

        # process properties
        for prop in vmconfig["state"]:
            # skip special vmconfig_types
            if (
                prop in vmconfig_type["instance"]
                or prop in vmconfig_type["collection"]
                or prop in vmconfig_type["create_only"]
            ):
                continue

            # skip unchanged properties
            if prop in vmconfig["current"]:
                if isinstance(vmconfig["current"][prop], (list)) or isinstance(
                    vmconfig["current"][prop], (dict)
                ):
                    if vmconfig["current"][prop] == vmconfig["state"][prop]:
                        continue
                else:
                    if "{}".format(vmconfig["current"][prop]) == "{}".format(
                        vmconfig["state"][prop]
                    ):
                        continue

            # add property to changeset
            vmconfig["changed"][prop] = vmconfig["state"][prop]

        # process collections
        for collection in vmconfig_type["collection"]:
            # skip create only collections
            if collection in vmconfig_type["create_only"]:
                continue

            # enforcement
            enforce = config[f"enforce_{collection}"]
            log.debug("smartos.vm_present::enforce_%s = %s", collection, enforce)

            # dockerinit handling
            if collection == "internal_metadata" and vmconfig["state"].get(
                "docker", False
            ):
                if "internal_metadata" not in vmconfig["state"]:
                    vmconfig["state"]["internal_metadata"] = {}

                # preserve some docker specific metadata (added and needed by dockerinit)
                for var in vmconfig_docker_keep:
                    val = vmconfig["current"].get(collection, {}).get(var, None)
                    if val is not None:
                        vmconfig["state"]["internal_metadata"][var] = val

            # process add and update for collection
            if (
                collection in vmconfig["state"]
                and vmconfig["state"][collection] is not None
            ):
                for prop in vmconfig["state"][collection]:
                    # skip unchanged properties
                    if (
                        prop in vmconfig["current"][collection]
                        and vmconfig["current"][collection][prop]
                        == vmconfig["state"][collection][prop]
                    ):
                        continue

                    # skip update if not enforcing
                    if not enforce and prop in vmconfig["current"][collection]:
                        continue

                    # create set_ dict
                    if f"set_{collection}" not in vmconfig["changed"]:
                        vmconfig["changed"][f"set_{collection}"] = {}

                    # add property to changeset
                    vmconfig["changed"][f"set_{collection}"][prop] = vmconfig["state"][
                        collection
                    ][prop]

            # process remove for collection
            if (
                enforce
                and collection in vmconfig["current"]
                and vmconfig["current"][collection] is not None
            ):
                for prop in vmconfig["current"][collection]:
                    # skip if exists in state
                    if (
                        collection in vmconfig["state"]
                        and vmconfig["state"][collection] is not None
                    ):
                        if prop in vmconfig["state"][collection]:
                            continue

                    # create remove_ array
                    if f"remove_{collection}" not in vmconfig["changed"]:
                        vmconfig["changed"][f"remove_{collection}"] = []

                    # remove property
                    vmconfig["changed"][f"remove_{collection}"].append(prop)

        # process instances
        for instance in vmconfig_type["instance"]:
            # skip create only instances
            if instance in vmconfig_type["create_only"]:
                continue

            # add or update instances
            if (
                instance in vmconfig["state"]
                and vmconfig["state"][instance] is not None
            ):
                for state_cfg in vmconfig["state"][instance]:
                    add_instance = True

                    # find instance with matching ids
                    for current_cfg in vmconfig["current"][instance]:
                        if vmconfig_type["instance"][instance] not in state_cfg:
                            continue

                        if (
                            state_cfg[vmconfig_type["instance"][instance]]
                            == current_cfg[vmconfig_type["instance"][instance]]
                        ):
                            # ids have matched, disable add instance
                            add_instance = False

                            changed = _get_instance_changes(current_cfg, state_cfg)
                            update_cfg = {}

                            # handle changes
                            for prop in changed:
                                update_cfg[prop] = state_cfg[prop]

                            # handle new properties
                            for prop in state_cfg:
                                # skip empty props like ips, options,..
                                if (
                                    isinstance(state_cfg[prop], (list))
                                    and not state_cfg[prop]
                                ):
                                    continue

                                if prop not in current_cfg:
                                    update_cfg[prop] = state_cfg[prop]

                            # update instance
                            if update_cfg:
                                # create update_ array
                                if f"update_{instance}" not in vmconfig["changed"]:
                                    vmconfig["changed"][f"update_{instance}"] = []

                                update_cfg[vmconfig_type["instance"][instance]] = (
                                    state_cfg[vmconfig_type["instance"][instance]]
                                )
                                vmconfig["changed"][f"update_{instance}"].append(
                                    update_cfg
                                )

                    if add_instance:
                        # create add_ array
                        if f"add_{instance}" not in vmconfig["changed"]:
                            vmconfig["changed"][f"add_{instance}"] = []

                        # add instance
                        vmconfig["changed"][f"add_{instance}"].append(state_cfg)

            # remove instances
            if (
                instance in vmconfig["current"]
                and vmconfig["current"][instance] is not None
            ):
                for current_cfg in vmconfig["current"][instance]:
                    remove_instance = True

                    # find instance with matching ids
                    if (
                        instance in vmconfig["state"]
                        and vmconfig["state"][instance] is not None
                    ):
                        for state_cfg in vmconfig["state"][instance]:
                            if vmconfig_type["instance"][instance] not in state_cfg:
                                continue

                            if (
                                state_cfg[vmconfig_type["instance"][instance]]
                                == current_cfg[vmconfig_type["instance"][instance]]
                            ):
                                # keep instance if matched
                                remove_instance = False

                    if remove_instance:
                        # create remove_ array
                        if f"remove_{instance}" not in vmconfig["changed"]:
                            vmconfig["changed"][f"remove_{instance}"] = []

                        # remove instance
                        vmconfig["changed"][f"remove_{instance}"].append(
                            current_cfg[vmconfig_type["instance"][instance]]
                        )

        # update vm if we have pending changes
        kvm_needs_start = False
        if not __opts__["test"] and vmconfig["changed"]:
            # stop kvm if disk updates and kvm_reboot
            if vmconfig["current"]["brand"] == "kvm" and config["kvm_reboot"]:
                if (
                    "add_disks" in vmconfig["changed"]
                    or "update_disks" in vmconfig["changed"]
                    or "remove_disks" in vmconfig["changed"]
                ):
                    if vmconfig["state"]["hostname"] in __salt__["vmadm.list"](
                        order="hostname", search="state=running"
                    ):
                        kvm_needs_start = True
                        __salt__["vmadm.stop"](
                            vm=vmconfig["state"]["hostname"], key="hostname"
                        )

            # do update
            rret = __salt__["vmadm.update"](
                vm=vmconfig["state"]["hostname"], key="hostname", **vmconfig["changed"]
            )
            if not isinstance(rret, (bool)) and "Error" in rret:
                ret["result"] = False
                ret["comment"] = "{}".format(rret["Error"])
            else:
                ret["result"] = True
                ret["changes"][vmconfig["state"]["hostname"]] = vmconfig["changed"]

        if ret["result"]:
            if __opts__["test"]:
                ret["changes"][vmconfig["state"]["hostname"]] = vmconfig["changed"]

            if (
                vmconfig["state"]["hostname"] in ret["changes"]
                and ret["changes"][vmconfig["state"]["hostname"]]
            ):
                ret["comment"] = "vm {} updated".format(vmconfig["state"]["hostname"])
                if (
                    config["kvm_reboot"]
                    and vmconfig["current"]["brand"] == "kvm"
                    and not __opts__["test"]
                ):
                    if vmconfig["state"]["hostname"] in __salt__["vmadm.list"](
                        order="hostname", search="state=running"
                    ):
                        __salt__["vmadm.reboot"](
                            vm=vmconfig["state"]["hostname"], key="hostname"
                        )
                    if kvm_needs_start:
                        __salt__["vmadm.start"](
                            vm=vmconfig["state"]["hostname"], key="hostname"
                        )
            else:
                ret["changes"] = {}
                ret["comment"] = "vm {} is up to date".format(
                    vmconfig["state"]["hostname"]
                )

            # reprovision (if required and allowed)
            if (
                "image_uuid" in vmconfig["current"]
                and vmconfig["reprovision_uuid"] != vmconfig["current"]["image_uuid"]
            ):
                if config["reprovision"]:
                    rret = __salt__["vmadm.reprovision"](
                        vm=vmconfig["state"]["hostname"],
                        key="hostname",
                        image=vmconfig["reprovision_uuid"],
                    )
                    if not isinstance(rret, (bool)) and "Error" in rret:
                        ret["result"] = False
                        ret["comment"] = "vm {} updated, reprovision failed".format(
                            vmconfig["state"]["hostname"]
                        )
                    else:
                        ret["comment"] = "vm {} updated and reprovisioned".format(
                            vmconfig["state"]["hostname"]
                        )
                        if vmconfig["state"]["hostname"] not in ret["changes"]:
                            ret["changes"][vmconfig["state"]["hostname"]] = {}
                        ret["changes"][vmconfig["state"]["hostname"]]["image_uuid"] = (
                            vmconfig["reprovision_uuid"]
                        )
                else:
                    log.warning(
                        "smartos.vm_present::%s::reprovision - "
                        "image_uuid in state does not match current, "
                        "reprovision not allowed",
                        name,
                    )
        else:
            ret["comment"] = "vm {} failed to be updated".format(
                vmconfig["state"]["hostname"]
            )
            if not isinstance(rret, (bool)) and "Error" in rret:
                ret["comment"] = "{}".format(rret["Error"])
    else:
        # check required image installed
        ret["result"] = True

        # disks need some special care
        if "disks" in vmconfig:
            new_disks = []
            for disk in vmconfig["disks"]:
                if "path" in disk:
                    del disk["path"]
                new_disks.append(disk)
            vmconfig["disks"] = new_disks

        # create vm
        if ret["result"]:
            uuid = (
                __salt__["vmadm.create"](**vmconfig) if not __opts__["test"] else True
            )
            if not isinstance(uuid, (bool)) and "Error" in uuid:
                ret["result"] = False
                ret["comment"] = "{}".format(uuid["Error"])
            else:
                ret["result"] = True
                ret["changes"][vmconfig["hostname"]] = vmconfig
                ret["comment"] = "vm {} created".format(vmconfig["hostname"])

    return ret


def vm_absent(name, archive=False):
    """
    Ensure vm is absent on the computenode

    name : string
        hostname of vm
    archive : boolean
        toggle archiving of vm on removal

    .. note::

        State ID is used as hostname. Hostnames must be unique.

    """
    name = name.lower()
    ret = {"name": name, "changes": {}, "result": None, "comment": ""}

    if name not in __salt__["vmadm.list"](order="hostname"):
        # we're good
        ret["result"] = True
        ret["comment"] = f"vm {name} is absent"
    else:
        # delete vm
        if not __opts__["test"]:
            # set archive to true if needed
            if archive:
                __salt__["vmadm.update"](
                    vm=name, key="hostname", archive_on_delete=True
                )

            ret["result"] = __salt__["vmadm.delete"](name, key="hostname")
        else:
            ret["result"] = True

        if not isinstance(ret["result"], bool) and ret["result"].get("Error"):
            ret["result"] = False
            ret["comment"] = f"failed to delete vm {name}"
        else:
            ret["comment"] = f"vm {name} deleted"
            ret["changes"][name] = None

    return ret


def vm_running(name):
    """
    Ensure vm is in the running state on the computenode

    name : string
        hostname of vm

    .. note::

        State ID is used as hostname. Hostnames must be unique.

    """
    name = name.lower()
    ret = {"name": name, "changes": {}, "result": None, "comment": ""}

    if name in __salt__["vmadm.list"](order="hostname", search="state=running"):
        # we're good
        ret["result"] = True
        ret["comment"] = f"vm {name} already running"
    else:
        # start the vm
        ret["result"] = (
            True if __opts__["test"] else __salt__["vmadm.start"](name, key="hostname")
        )
        if not isinstance(ret["result"], bool) and ret["result"].get("Error"):
            ret["result"] = False
            ret["comment"] = f"failed to start {name}"
        else:
            ret["changes"][name] = "running"
            ret["comment"] = f"vm {name} started"

    return ret


def vm_stopped(name):
    """
    Ensure vm is in the stopped state on the computenode

    name : string
        hostname of vm

    .. note::

        State ID is used as hostname. Hostnames must be unique.

    """
    name = name.lower()
    ret = {"name": name, "changes": {}, "result": None, "comment": ""}

    if name in __salt__["vmadm.list"](order="hostname", search="state=stopped"):
        # we're good
        ret["result"] = True
        ret["comment"] = f"vm {name} already stopped"
    else:
        # stop the vm
        ret["result"] = (
            True if __opts__["test"] else __salt__["vmadm.stop"](name, key="hostname")
        )
        if not isinstance(ret["result"], bool) and ret["result"].get("Error"):
            ret["result"] = False
            ret["comment"] = f"failed to stop {name}"
        else:
            ret["changes"][name] = "stopped"
            ret["comment"] = f"vm {name} stopped"

    return ret
